# Copyright 2024-2025, Battelle Energy Alliance, LLC All Rights Reserved.
import re

from montepy.errors import UnsupportedFeature

"""
Constants related to how MCNP inputs are formatted, and MontePy behavior.
"""

rel_tol = 1e-9
"""Relative tolerance passed to math.isclose.

Returns
-------
float
"""

abs_tol = 0.0
"""Absolute tolerance passed to math.isclose.

Returns
-------
float
"""


BLANK_SPACE_CONTINUE = 5
"""Number of spaces in a new line before it's considered a continuation."""

COMMENT_FINDER = re.compile(rf"\s{{0,{BLANK_SPACE_CONTINUE - 1}}}c", re.IGNORECASE)
"""A regular expression for finding the start of a ``c`` style comment."""

LINE_LENGTH = {
    (5, 1, 60): 80,
    (6, 1, 0): 80,
    (6, 2, 0): 128,
    (6, 3, 0): 128,
    (6, 3, 1): 128,
}
"""The number of characters allowed in a line for each MCNP version.

Citations:

* 5.1.60 and 6.1.0: Section 2.6.2 of `LA-UR-18-20808 <https://www.osti.gov/biblio/1419730>`_
* 6.2.0: Section 1.1.1: :manual62:`13`
* 6.3.0: :manual63:`3.2.2`
* 6.3.1: Section 3.2.2 of `LA-UR-24-24602 <https://www.osti.gov/biblio/2372634>`_
"""

DEFAULT_VERSION = (6, 3, 0)
"""The default version of MCNP to use."""

TABSIZE = 8
"""How many spaces a tab is expand to."""

ASCII_CEILING = 127
"""The maximum allowed code point allowed by ASCII.

Source: `Wikipedia <https://en.wikipedia.org/wiki/ASCII>`_
"""

MAX_ATOMIC_SYMBOL_LENGTH = 2
"""The maximum length of an atomic symbol."""


def get_max_line_length(mcnp_version=DEFAULT_VERSION):
    """Gets the maximum allowed length for an input line for a specific MCNP version.

    The version must be a three component tuple e.g., (6, 2, 0) and (5, 1, 60).
    Line lengths inferred from:

    | C. J. Werner et al., “MCNP Version 6.2 Release Notes,” LA-UR--18-20808, 1419730, Feb. 2018. doi: 10.2172/1419730. section 2.6.2

    Prior MCNP release version numbers were taken from `RSICC <https://rsicc.ornl.gov/codes/ccc/ccc8/ccc-810.html>`_.

    Parameters
    ----------
    mcnp_version : tuple
        The version of MCNP that the input is intended for.

    Returns
    -------
    int
        The number of characters allowed in a line.
    """
    if mcnp_version >= DEFAULT_VERSION:
        return LINE_LENGTH[DEFAULT_VERSION]
    try:
        return LINE_LENGTH[mcnp_version]
    except KeyError:
        raise UnsupportedFeature(f"MCNP Version {mcnp_version} is not supported.")



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.

import traceback


class LineOverRunWarning(UserWarning):
    """Raised when non-comment inputs exceed the allowed line length in an input."""

    def __init__(self, message):
        self.message = message


class MalformedInputError(ValueError):
    """Raised when there is an error with the MCNP input not related to the parser."""

    def __init__(self, input, message):
        if input and getattr(input, "input_file", None) and input.input_file:
            self.input = input
            path = input.input_file.path
            start_line = input.line_number
            self.path = path
            self.start = start_line
            lines = "\n".join(input.input_lines)
        else:
            path = ""
            start_line = 0
            lines = ""
        self.message = message
        super().__init__(self.message)


class ParsingError(MalformedInputError):
    """Raised when there is an error parsing the MCNP input at the SLY parsing layer."""

    def __init__(self, input, message, error_queue):
        messages = []
        if input and getattr(input, "input_file", None) and input.input_file:
            self.input = input
            path = input.input_file.path
            start_line = input.line_number
            self.path = path
            self.start = start_line
        else:
            path = ""
            start_line = 0
        if error_queue:
            for error in error_queue:
                if token := error["token"]:
                    line_no = error["line"]
                    index = error["index"]
                    self.rel_line = line_no
                    self.abs_line = line_no + start_line
                    base_message = f'There was an error parsing "{token.value}".'
                else:
                    line_no = 0
                    index = 0
                    base_message = f"The input ended prematurely."
                messages.append(
                    _print_input(
                        path,
                        start_line,
                        error["message"],
                        line_no,
                        input,
                        token,
                        base_message,
                        index,
                    )
                )
            self.message = "\n".join(messages + [message])
        else:
            self.message = message

        ValueError.__init__(self, self.message)


def _print_input(
    path,
    start_line,
    error_msg,
    line_no=0,
    input=None,
    token=None,
    base_message=None,
    index=None,
):
    buffer = [f"    {path}, line {start_line + line_no -1}", ""]
    if input:
        for i, line in enumerate(input.input_lines):
            if i == line_no - 1:
                buffer.append(f"    >{start_line + i:5g}| {line}")
                if token:
                    length = len(token.value)
                    marker = "^" * length
                    buffer.append(
                        f"{' '* 10}|{' ' * (index+1)}{marker} not expected here."
                    )
            else:
                buffer.append(f"     {start_line + i:5g}| {line}")
        if base_message:
            buffer.append(base_message)
        buffer.append(error_msg)
    return "\n".join(buffer)


class NumberConflictError(Exception):
    """Raised when there is a conflict in number spaces"""

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class BrokenObjectLinkError(MalformedInputError):
    """Raised when the referenced object does not exist in the input file."""

    def __init__(self, parent_type, parent_number, child_type, child_number):
        """
        Parameters
        ----------
        parent_type : str
            Name of the parent object linking (e.g., Cell)
        parent_number : int
            the number of the parent object
        child_type : str
            Name of the type of object missing in the link (e.g.,
            surface)
        child_number : int
            the number for the missing object
        """
        super().__init__(
            None,
            f"{child_type} {child_number} is missing from the input from the definition of: {parent_type} {parent_number}",
        )


class RedundantParameterSpecification(ValueError):
    """Raised when multiple conflicting parameters are given.

    e.g., ``1 0 -1 imp:n=5 imp:n=0``
    """

    def __init__(self, key, new_value):
        self.message = (
            f"Multiple values given for parameter: {key}. new_value given: {new_value}."
        )
        super().__init__(self.message)


class ParticleTypeWarning(Warning):
    """Base class for incongruencies between particle types
    in problem mode, cell importances, and tallies
    """

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class ParticleTypeNotInProblem(ParticleTypeWarning):
    """Raised when data, such as cell importance or tally type,
    are set for a particle type not in the problem's mode.
    """

    pass


class ParticleTypeNotInCell(ParticleTypeWarning):
    """Raised when data for importance data for a particle in
    the problem is not provided for a cell.
    """

    pass


class UnsupportedFeature(NotImplementedError):
    """Raised when MCNP syntax that is not supported is found"""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class UnknownElement(ValueError):
    """Raised when an undefined element is used."""

    def __init__(self, missing_val):
        self.message = f"An element identified by: {missing_val} is unknown to MontePy."
        super().__init__(self.message)


class IllegalState(ValueError):
    """Raised when an object can't be printed out due to an illegal state."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class LineExpansionWarning(Warning):
    """Warning for when a field or line expands that may damage user formatting."""

    pass


def add_line_number_to_exception(error, broken_robot):
    """Adds additional context to an Exception raised by an :class:`~montepy.mcnp_object.MCNP_Object`.

    This will add the line, file name, and the input lines to the error.

    Parameters
    ----------
    error : Exception
        The error that was raised.
    broken_robot : MCNP_Object
        The parent object that had the error raised.

    Raises
    ------
    Exception
        ... that's the whole point.
    """
    # avoid calling this n times recursively
    if hasattr(error, "montepy_handled"):
        raise error
    error.montepy_handled = True
    args = error.args
    trace = error.__traceback__
    if len(args) > 0:
        message = args[0]
    else:
        message = ""
    try:
        input_obj = broken_robot._input
        assert input_obj is not None
        lineno = input_obj.line_number
        file = str(input_obj.input_file)
        lines = input_obj.input_lines
        message = _print_input(file, lineno, message, input=input_obj)
    except Exception as e:
        try:
            message = (
                f"{message}\n\nError came from {broken_robot} from an unknown file."
            )
        except Exception as e2:
            message = f"{message}\n\nError came from an object of type {type(broken_robot)} from an unknown file."
    args = (message,) + args[1:]
    error.args = args
    raise error.with_traceback(trace)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools

from montepy.input_parser.data_parser import DataParser
from montepy.input_parser import syntax_node


class MaterialParser(DataParser):
    debugfile = None

    @_(
        "classifier_phrase mat_data",
        "padding classifier_phrase mat_data",
    )
    def material(self, p):
        ret = {}
        if isinstance(p[0], syntax_node.PaddingNode):
            ret["start_pad"] = p.padding
        else:
            ret["start_pad"] = syntax_node.PaddingNode()
        ret["classifier"] = p.classifier_phrase
        ret["data"] = p.mat_data
        return syntax_node.SyntaxNode("data", ret)

    @_("mat_datum", "mat_data mat_datum")
    def mat_data(self, p):
        if len(p) == 1:
            ret = syntax_node.MaterialsNode("mat stuff")
        else:
            ret = p.mat_data
        datum = p.mat_datum
        if isinstance(datum, tuple):
            ret.append_nuclide(datum)
        elif isinstance(datum, syntax_node.ListNode):
            [ret.append_nuclide(n) for n in self._convert_to_isotope(datum)]
        else:
            ret.append_param(datum)
        return ret

    def _convert_to_isotope(self, old):
        new_list = []

        def batch_gen():
            it = iter(old)
            while batch := tuple(itertools.islice(it, 2)):
                yield batch

        for group in batch_gen():
            if group[0].type != str:
                group[0]._convert_to_str()
            new_list.append(("foo", *group))
        return new_list

    @_("isotope_fraction", "even_number_sequence", "parameter", "mat_parameter")
    def mat_datum(self, p):
        return p[0]

    @_(
        "classifier param_seperator library",
    )
    def mat_parameter(self, p):
        """A singular Key-value pair that includes a material library.

        Returns
        -------
        SyntaxNode
            the parameter.
        """
        return syntax_node.SyntaxNode(
            p.classifier.prefix.value,
            {"classifier": p.classifier, "seperator": p.param_seperator, "data": p[2]},
        )

    @_("NUMBER_WORD")
    @_("NUMBER_WORD padding")
    def library(self, p):
        """A library name."""
        return self._flush_phrase(p, str)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from enum import Enum
import itertools
import os
import warnings

from montepy.data_inputs import mode, transform
from montepy._cell_data_control import CellDataPrintController
from montepy.cell import Cell
from montepy.cells import Cells
from montepy.errors import *
from montepy.constants import DEFAULT_VERSION
from montepy.materials import Material, Materials
from montepy.surfaces import surface, surface_builder
from montepy.surface_collection import Surfaces

# weird way to avoid circular imports
from montepy.data_inputs import parse_data
from montepy.input_parser import input_syntax_reader, block_type, mcnp_input
from montepy.input_parser.input_file import MCNP_InputFile
from montepy.universes import Universe, Universes
from montepy.transforms import Transforms
import montepy


class MCNP_Problem:
    """A class to represent an entire MCNP problem in a semantic way.

    Notes
    -----
    If a stream is provided. It will not be closed by this function.

    Parameters
    ----------
    destination : io.TextIOBase, str, os.PathLike
        the path to the input file to read, or a readable stream.
    """

    _NUMBERED_OBJ_MAP = {
        Cell: Cells,
        surface.Surface: Surfaces,
        Material: Materials,
        transform.Transform: Transforms,
        Universe: Universes,
    }

    def __init__(self, destination):
        if hasattr(destination, "read") and callable(getattr(destination, "read")):
            self._input_file = MCNP_InputFile.from_open_stream(destination)
        elif isinstance(destination, (str, os.PathLike)):
            self._input_file = MCNP_InputFile(destination)
        self._title = None
        self._message = None
        self.__unpickled = False
        self._print_in_data_block = CellDataPrintController()
        self._original_inputs = []
        for collect_type in self._NUMBERED_OBJ_MAP.values():
            attr_name = f"_{collect_type.__name__.lower()}"
            setattr(self, attr_name, collect_type(problem=self))
        self._data_inputs = []
        self._mcnp_version = DEFAULT_VERSION
        self._mode = mode.Mode()

    def __setstate__(self, nom_nom):
        self.__dict__.update(nom_nom)
        self.__unpickled = True

    @staticmethod
    def __get_collect_attr_name(collect_type):
        return f"_{collect_type.__name__.lower()}"

    @property
    def original_inputs(self):
        """A list of the MCNP_Inputs read from the original file.

        This should not be mutated, and should be used as a reference to maintain
        the structure

        .. deprecated:: 0.2.0
            This will likely be removed soon, and it's functionality will not be necessary to reproduce.

        Returns
        -------
        list
            A list of the MCNP_Object objects representing the file as
            it was read
        """
        return self._original_inputs

    def __relink_objs(self):
        if self.__unpickled:
            for collection in set(self._NUMBERED_OBJ_MAP.values()) | {"_data_inputs"}:
                if not isinstance(collection, str):
                    collection = self.__get_collect_attr_name(collection)
                collection = getattr(self, collection)
                if isinstance(
                    collection,
                    montepy.numbered_object_collection.NumberedObjectCollection,
                ):
                    collection.link_to_problem(self)
                else:
                    for obj in collection:
                        obj.link_to_problem(self)
            self.__unpickled = False

    def __unlink_objs(self):
        for collection in set(self._NUMBERED_OBJ_MAP.values()) | {"_data_inputs"}:
            if not isinstance(collection, str):
                collection = self.__get_collect_attr_name(collection)
            collection = getattr(self, collection)
            if isinstance(
                collection,
                montepy.numbered_object_collection.NumberedObjectCollection,
            ):
                collection.link_to_problem(None)
            else:
                for obj in collection:
                    obj.link_to_problem(None)
        self.__unpickled = True
        self.__relink_objs()

    def __deepcopy__(self, memo):
        cls = type(self)
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        result.__unlink_objs()
        return result

    def clone(self):
        """Creates a complete independent copy of this problem.

        .. versionadded:: 0.5.0

        Returns
        -------
        MCNP_Problem
        """
        return copy.deepcopy(self)

    @property
    def cells(self):
        """A collection of the Cell objects in this problem.

        Returns
        -------
        Cells
            a collection of the Cell objects, ordered by the order they
            were in the input file.
        """
        self.__relink_objs()
        return self._cells

    @cells.setter
    def cells(self, cells):
        if not isinstance(cells, (Cells, list)):
            raise TypeError("cells must be an instance of list or Cells")
        if isinstance(cells, list):
            cells = Cells(cells)
        if cells is self.cells:
            return
        self.cells.clear()
        self.cells.extend(cells)

    @property
    def mode(self):
        """The mode of particles being used for the problem.

        Returns
        -------
        Mode
        """
        return self._mode

    def set_mode(self, particles):
        """Sets the mode of problem to the given particles.

        For details see: :func:`montepy.data_cards.mode.Mode.set`.

        Parameters
        ----------
        particles : list, str
            the particles that the mode will be switched to.

        Raises
        ------
        ValueError
            if string is not a valid particle shorthand.
        """
        self._mode.set(particles)

    @property
    def mcnp_version(self):
        """The version of MCNP that this is intended for.

        Notes
        -----
        MCNP versions prior to 6.2 aren't fully supported to avoid
        Export Control Restrictions. Documentation for MCNP 6.2 is public in report:
        LA-UR-17-29981.
        All features are based on MCNP 6.2, and may cause other versions of MCNP to break.


        The version is a tuple of major, minor, revision.
        6.2.0 would be represented as (6, 2, 0)

        Returns
        -------
        tuple
        """
        return self._mcnp_version

    @mcnp_version.setter
    def mcnp_version(self, version):
        """
        Parameters
        ----------
        version : tuple
            the version tuple. Must be greater than 6.2.0
        """
        if version < (5, 1, 60):
            raise ValueError(f"The mcnp_version {version} is not supported by MontePy")
        self._mcnp_version = version

    @property
    def surfaces(self):
        """A collection of the Surface objects in this problem.

        Returns
        -------
        Surfaces
            a collection of the Surface objects, ordered by the order
            they were in the input file.
        """
        self.__relink_objs()
        return self._surfaces

    @surfaces.setter
    def surfaces(self, surfs):
        if not isinstance(surfs, (list, Surfaces)):
            raise TypeError("Surfaces must be of type list or Surfaces")
        if isinstance(surfs, list):
            surfs = Surfaces(surfs)
        surfs.link_to_problem(self)
        self._surfaces = surfs

    @property
    def materials(self):
        """A collection of the Material objects in this problem.

        Returns
        -------
        Materials
            a colection of the Material objects, ordered by the order
            they were in the input file.
        """
        self.__relink_objs()
        return self._materials

    @materials.setter
    def materials(self, mats):
        if not isinstance(mats, (list, Materials)):
            raise TypeError("materials must be of type list and Materials")
        if isinstance(mats, list):
            mats = Materials(mats)
        mats.link_to_problem(self)
        self._materials = mats

    @property
    def print_in_data_block(self):
        """Controls whether or not the specific input gets printed in the cell block or the data block.

        This acts like a dictionary. The key is the case insensitive name of the card.
        For example to enable printing importance data in the data block run:

        ``problem.print_in_data_block["Imp"] = True``

        Returns
        -------
        bool
        """
        return self._print_in_data_block

    @property
    def data_inputs(self):
        """A list of the DataInput objects in this problem.

        Returns
        -------
        list
            a list of the
            :class:`~montepy.data_cards.data_card.DataCardAbstract`
            objects, ordered by the order they were in the input file.
        """
        self.__relink_objs()
        return self._data_inputs

    @property
    def input_file(self):
        """The file name of the original file name this problem was read from.

        Returns
        -------
        MCNP_InputFile
        """
        return self._input_file

    @property
    def message(self):
        """The Message object at the beginning of the problem if any.

        Returns
        -------
        Message
        """
        return self._message

    @property
    def title(self):
        """The Title object for the title.

        Returns
        -------
        Title
        """
        return self._title

    @title.setter
    def title(self, title):
        """
        Parameters
        ----------
        title : The str for the title to be set to.
        """
        self._title = mcnp_input.Title([title], title)

    @property
    def universes(self):
        """The Universes object holding all problem universes.

        Returns
        -------
        Universes
            a collection of universes in the problem.
        """
        return self._universes

    @property
    def transforms(self):
        """The collection of transform objects in this problem.

        Returns
        -------
        Transforms
            a collection of transforms in the problem.
        """
        return self._transforms

    def parse_input(self, check_input=False, replace=True):
        """Semantically parses the MCNP file provided to the constructor.

        Parameters
        ----------
        check_input : bool
            If true, will try to find all errors with input and collect
            them as warnings to log.
        replace : bool
            replace all non-ASCII characters with a space (0x20)
        """
        trailing_comment = None
        last_obj = None
        last_block = None
        OBJ_MATCHER = {
            block_type.BlockType.CELL: (Cell, self._cells),
            block_type.BlockType.SURFACE: (
                surface_builder.parse_surface,
                self._surfaces,
            ),
            block_type.BlockType.DATA: (parse_data, self._data_inputs),
        }
        try:
            for i, input in enumerate(
                input_syntax_reader.read_input_syntax(
                    self._input_file, self.mcnp_version, replace=replace
                )
            ):
                self._original_inputs.append(input)
                if i == 0 and isinstance(input, mcnp_input.Message):
                    self._message = input

                elif isinstance(input, mcnp_input.Title) and self._title is None:
                    self._title = input

                elif isinstance(input, mcnp_input.Input):
                    if last_block != input.block_type:
                        trailing_comment = None
                        last_block = input.block_type
                    obj_parser, obj_container = OBJ_MATCHER[input.block_type]
                    if len(input.input_lines) > 0:
                        try:
                            obj = obj_parser(input)
                            obj.link_to_problem(self)
                            if isinstance(
                                obj_container,
                                montepy.numbered_object_collection.NumberedObjectCollection,
                            ):
                                obj_container.append(obj, initial_load=True)
                            else:
                                obj_container.append(obj)
                        except (
                            MalformedInputError,
                            NumberConflictError,
                            ParsingError,
                            UnknownElement,
                        ) as e:
                            if check_input:
                                warnings.warn(
                                    f"{type(e).__name__}: {e.message}", stacklevel=2
                                )
                                continue
                            else:
                                raise e
                        if isinstance(obj, Material):
                            self._materials.append(obj, insert_in_data=False)
                        if isinstance(obj, transform.Transform):
                            self._transforms.append(obj, insert_in_data=False)
                    if trailing_comment is not None and last_obj is not None:
                        obj._grab_beginning_comment(trailing_comment, last_obj)
                        last_obj._delete_trailing_comment()
                    trailing_comment = obj.trailing_comment
                    last_obj = obj
        except UnsupportedFeature as e:
            if check_input:
                warnings.warn(f"{type(e).__name__}: {e.message}", stacklevel=2)
            else:
                raise e
        self.__update_internal_pointers(check_input)

    def __update_internal_pointers(self, check_input=False):
        """Updates the internal pointers between objects

        Parameters
        ----------
        check_input : bool
            If true, will try to find all errors with input and collect
            them as warnings to log.
        """

        def handle_error(e):
            if check_input:
                warnings.warn(f"{type(e).__name__}: {e.message}", stacklevel=3)
            else:
                raise e

        self.__load_data_inputs_to_object(self._data_inputs)
        self._cells.update_pointers(
            self.cells,
            self.materials,
            self.surfaces,
            self._data_inputs,
            self,
            check_input,
        )
        for surface in self._surfaces:
            try:
                surface.update_pointers(self.surfaces, self._data_inputs)
            except (BrokenObjectLinkError,) as e:
                handle_error(e)
        to_delete = []
        for data_index, data_input in enumerate(self._data_inputs):
            try:
                if data_input.update_pointers(self._data_inputs):
                    to_delete.append(data_index)
            except (
                BrokenObjectLinkError,
                MalformedInputError,
            ) as e:
                handle_error(e)
                continue
        for delete_index in to_delete[::-1]:
            del self._data_inputs[delete_index]

    def remove_duplicate_surfaces(self, tolerance):
        """Finds duplicate surfaces in the problem, and remove them.

        Parameters
        ----------
        tolerance : float
            The amount of relative error to consider two surfaces
            identical
        """
        to_delete = montepy.surface_collection.Surfaces()
        matching_map = {}
        for surface in self.surfaces:
            if surface not in to_delete:
                matches = surface.find_duplicate_surfaces(self.surfaces, tolerance)
                if matches:
                    for match in matches:
                        to_delete.add(match)
                        matching_map[match.number] = (match, surface)
        for cell in self.cells:
            cell.remove_duplicate_surfaces(matching_map)
        self.__update_internal_pointers()
        for surface in to_delete:
            self._surfaces.remove(surface)

    def add_cell_children_to_problem(self):  # pragma: no cover
        """Deprecated: Adds the surfaces, materials, and transforms of all cells in this problem to this problem to the
           internal lists to allow them to be written to file.

        .. deprecated:: 1.0.0

            This function is no longer needed. When cells are added to problem.cells these children are added as well.

        Raises
        ------
        DeprecationWarning
        """
        raise DeprecationWarning(
            "add_cell_children_to_problem has been removed,"
            " as the children are automatically added with the cell."
        )

    def write_problem(self, destination, overwrite=False):
        """Write the problem to a file or writeable object.

        Parameters
        ----------
        destination : io.TextIOBase, str, os.PathLike
            File path or writable object
        overwrite : bool
            Whether to overwrite 'destination' if it is an existing file
        """
        if hasattr(destination, "write") and callable(getattr(destination, "write")):
            new_file = MCNP_InputFile.from_open_stream(destination)
            self._write_to_stream(new_file)
        elif isinstance(destination, (str, os.PathLike)):
            new_file = MCNP_InputFile(destination, overwrite=overwrite)
            with new_file.open("w") as fh:
                self._write_to_stream(fh)
        else:
            raise TypeError(
                f"destination f{destination} is not a file path or writable object"
            )

    def write_to_file(self, file_path, overwrite=False):
        """Writes the problem to a file.

        .. versionchanged:: 0.3.0
            The overwrite parameter was added.

        Parameters
        ----------
        file_path : str, os.PathLike
            the file path to write this problem to
        overwrite : bool
            Whether to overwrite the file at 'new_problem' if it exists

        Raises
        ------
        IllegalState
            if an object in the problem has not been fully initialized.
        FileExistsError
            if a file already exists with the same path.
        IsADirectoryError
            if the path given is actually a directory.
        """
        return self.write_problem(file_path, overwrite)

    def _write_to_stream(self, inp):
        """Writes the problem to a writeable stream.

        Parameters
        ----------
        inp : MCNP_InputFile
            Writable input file
        """
        with warnings.catch_warnings(record=True) as warning_catch:
            objects_list = []
            if self.message:
                objects_list.append(([self.message], False))
            objects_list += [
                ([self.title], False),
                (self.cells, True),
                (self.surfaces, True),
                (self.data_inputs, True),
            ]
            for objects, terminate in objects_list:
                for obj in objects:
                    lines = obj.format_for_mcnp_input(self.mcnp_version)
                    if warning_catch:
                        # handle ALL new warnings
                        for warning in warning_catch[::-1]:
                            if getattr(warning, "handled", None):
                                break
                            warning.lineno = inp.lineno
                            warning.path = inp.name
                            warning.obj = obj
                            warning.lines = lines
                            warning.handled = True
                    for line in lines:
                        inp.write(line + "\n")
                if terminate:
                    inp.write("\n")
            for line in self.cells._run_children_format_for_mcnp(
                self.data_inputs, self.mcnp_version
            ):
                inp.write(line + "\n")

            inp.write("\n")
        self._handle_warnings(warning_catch)

    def _handle_warnings(self, warning_queue):
        class WarningLevels(Enum):
            SUPRESS = 0
            MINIMAL = 1
            MAXIMAL = 5

        warning_level = WarningLevels.MAXIMAL

        for warning_message in warning_queue:
            warning = warning_message.message
            message = f"The input starting on Line {warning_message.lineno} of: {warning_message.path} expanded. "
            if warning_level == WarningLevels.SUPRESS:
                continue
            elif warning_level == WarningLevels.MAXIMAL:
                message += "\nThe new input is:\n"
                width = 15
                for i, line in enumerate(warning_message.lines):
                    message += f"     {warning_message.lineno + i:5g}| {line}\n"
                if hasattr(warning, "olds"):
                    message += (
                        f"\n    {'old values': ^{width}s} {'new values': ^{width}s}"
                    )
                    message += f"\n    {'':-^{width}s} {'':-^{width}s}\n"
                    for old, new in zip(warning.olds, warning.news):
                        formatter = f"    {{old: >{width}}} {{new: >{width}}}\n"
                        message += formatter.format(old=old, new=new)

            warning = LineExpansionWarning(message)
            warnings.warn(warning, stacklevel=3)

    def __load_data_inputs_to_object(self, data_inputs):
        """Loads data input into their appropriate problem attribute.

        Problem-level input should be loaded this way like: mode and kcode.
        """
        inputs_to_property = {mode.Mode: "_mode"}
        inputs_loaded = set()
        for input in data_inputs:
            if type(input) in inputs_to_property:
                if type(input) in inputs_loaded:
                    raise MalformedInputError(
                        input,
                        f"The input: {type(input)} is only allowed once in a problem",
                    )
                setattr(self, inputs_to_property[type(input)], input)
                inputs_loaded.add(type(input))

    def __str__(self):
        return f"MCNP problem for: {self._input_file}, {self._title}"

    def __repr__(self):
        ret = f"MCNP problem for: {self._input_file}\n"
        if self.message:
            ret += str(self._message) + "\n"
        ret += str(self._title) + "\n"
        for collection in [self.cells, self.surfaces, self.data_inputs]:
            for obj in collection:
                ret += f"{obj}\n"
            ret += "\n"
        return ret

    def parse(self, input: str, append: bool = True) -> montepy.mcnp_object.MCNP_Object:
        """Parses the MCNP object given by the string, and links it adds it to this problem.

        This attempts to identify the input type by trying to parse it in the following order:

        #. Data Input
        #. Surface
        #. Cell

        This is done mostly for optimization to go from easiest parsing to hardest.
        This will:

        #. Parse the input
        #. Link it to other objects in the problem. Note: this will raise an error if those objects don't exist.
        #. Append it to the appropriate collection

        Parameters
        ----------
        input : str
            the string describing the input. New lines are allowed but
            this does not need to meet MCNP line length rules.
        append : bool
            Whether to append this parsed object to this problem.

        Returns
        -------
        MCNP_Object
            the parsed object.

        Raises
        ------
        TypeError
            If a str is not given
        ParsingError
            If this is not a valid input.
        BrokenObjectLinkError
            if the dependent objects are not already in the problem.
        NumberConflictError
            if the object's number is already taken
        """
        try:
            obj = montepy.parse_data(input)
        except ParsingError:
            try:
                obj = montepy.parse_surface(input)
            except ParsingError:
                obj = montepy.Cell(input)
                # let final parsing error bubble up
        obj.link_to_problem(self)
        if isinstance(obj, montepy.Cell):
            obj.update_pointers(self.cells, self.materials, self.surfaces)
            if append:
                self.cells.append(obj)
        elif isinstance(obj, montepy.surfaces.surface.Surface):
            obj.update_pointers(self.surfaces, self.data_inputs)
            if append:
                self.surfaces.append(obj)
        else:
            obj.update_pointers(self.data_inputs)
            if append:
                self.data_inputs.append(obj)
                if isinstance(obj, Material):
                    self._materials.append(obj, insert_in_data=False)
                if isinstance(obj, transform.Transform):
                    self._transforms.append(obj, insert_in_data=False)
        return obj



# Copyright 2025, Battelle Energy Alliance, LLC All Rights Reserved.

import montepy
import pytest
import numpy as np


def test_cell_integral():
    montepy.Cell(number=3)
    montepy.Cell(number=np.uint8(1))
    with pytest.raises(TypeError):
        montepy.Cell(number=5.0)


def test_cell_real():
    c = montepy.Cell()
    c.atom_density = 1
    c.mass_density = np.float32(1.2e-3)


def test_surf_coeff_real():
    cx = montepy.CylinderParAxis()
    cx.coordinates = (0, np.int16(-1.23))



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.cell import Cell
from montepy.particle import Particle
from montepy.data_inputs.importance import Importance
from montepy.errors import *
from montepy.input_parser import mcnp_input, block_type
import os
import io
import pytest


def test_importance_init_cell():
    # test_normal cell init
    in_str = "1 0 -1 IMP:N,P=1"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    cell = Cell(card)
    assert cell.importance.neutron == 1.0
    assert cell.importance.photon == 1.0
    assert cell.importance.alpha_particle == 0.0
    assert cell.importance.all is None
    assert cell.importance.in_cell_block
    # test non-number imp
    in_str = "1 0 -1 IMP:N,P=h"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(ValueError):
        cell = Cell(card)
    # test negative imp
    in_str = "1 0 -1 IMP:N,P=-2"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(ValueError):
        cell = Cell(card)


def test_importance_init_data():
    in_str = "IMP:N,P 1 0"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    imp = Importance(card)
    assert [
        val.value for val in imp._particle_importances[Particle.NEUTRON]["data"]
    ] == [1.0, 0.0]
    assert [
        val.value for val in imp._particle_importances[Particle.PHOTON]["data"]
    ] == [1.0, 0.0]
    # test non-number imp
    in_str = "IMP:N,P 1 h"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(ValueError):
        imp = Importance(card)
    # test negative
    in_str = "IMP:N,P 1 -2"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(ValueError):
        imp = Importance(card)
    # test bad in_cell_block
    in_str = "IMP:N,P 1 2"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(TypeError):
        imp = Importance(card, in_cell_block=1)
    # test bad key
    in_str = "IMP:N,P 1 2"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(TypeError):
        imp = Importance(card, key=1)
    # test bad value
    in_str = "IMP:N,P 1 2"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    with pytest.raises(TypeError):
        imp = Importance(card, value=1)


def test_importance_iter_getter_in():
    in_str = "1 0 -1 IMP:N,P=1"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    cell = Cell(card)
    imp = cell.importance
    particles = [
        montepy.particle.Particle.NEUTRON,
        montepy.particle.Particle.PHOTON,
    ]
    for particle in imp:
        assert particle in particles
        assert imp[particle] == 1.0
    for particle in particles:
        assert particle in imp
    with pytest.raises(TypeError):
        imp["hi"]


def test_importance_all_setter():
    in_str = "1 0 -1 IMP:N,P=1"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    cell = Cell(card)
    problem = montepy.mcnp_problem.MCNP_Problem("foo")
    problem.mode.add(montepy.particle.Particle.NEUTRON)
    problem.mode.add(montepy.particle.Particle.PHOTON)
    imp = cell.importance
    cell.link_to_problem(problem)
    imp.all = 2.0
    assert imp.neutron == 2.0
    assert imp.photon == 2.0
    # try wrong type
    with pytest.raises(TypeError):
        imp.all = "h"
    # try negative type
    with pytest.raises(ValueError):
        imp.all = -2.0


def test_importance_setter():
    in_str = "1 0 -1 IMP:N,P=1"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    cell = Cell(card)
    cell.importance.neutron = 2.5
    assert cell.importance.neutron == 2.5
    problem = montepy.mcnp_problem.MCNP_Problem("foo")
    cell.link_to_problem(problem)
    # test problem mode enforcement
    with pytest.warns(ParticleTypeNotInProblem):
        cell.importance.photon = 1.0
    # test wrong type
    with pytest.raises(TypeError):
        cell.importance.neutron = "h"
    # test negative
    with pytest.raises(ValueError):
        cell.importance.neutron = -0.5

    cell.importance[Particle.NEUTRON] = 3
    assert cell.importance.neutron == 3.0
    with pytest.raises(TypeError):
        cell.importance[""] = 5
    with pytest.raises(TypeError):
        cell.importance[Particle.NEUTRON] = ""
    with pytest.raises(ValueError):
        cell.importance[Particle.NEUTRON] = -1.0


def test_importance_deleter():
    in_str = "1 0 -1 IMP:N,P=1"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    cell = Cell(card)
    del cell.importance.neutron
    assert cell.importance.neutron == 0.0
    del cell.importance[Particle.PHOTON]
    assert cell.importance.photon == 0.0
    with pytest.raises(TypeError):
        del cell.importance[""]


def test_importance_merge():
    in_str = "IMP:N,P 1 0"
    card = mcnp_input.Input([in_str], block_type.BlockType.DATA)
    imp1 = Importance(card)
    in_str = "IMP:E 0 0"
    card = mcnp_input.Input([in_str], block_type.BlockType.DATA)
    imp2 = Importance(card)
    imp1.merge(imp2)
    assert [
        val.value for val in imp1._particle_importances[Particle.NEUTRON]["data"]
    ] == [1.0, 0.0]
    assert [
        val.value for val in imp1._particle_importances[Particle.ELECTRON]["data"]
    ] == [0.0, 0.0]
    # test bad type
    with pytest.raises(TypeError):
        imp1.merge("hi")
    # test bad block type
    in_str = "1 0 -1 IMP:N,P=1"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    cell = Cell(card)
    with pytest.raises(ValueError):
        imp1.merge(cell.importance)
    in_str = "IMP:P 0 0"
    card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
    imp2 = Importance(card)
    with pytest.raises(MalformedInputError):
        imp1.merge(imp2)


def test_redundant_importance():
    with pytest.raises(MalformedInputError):
        montepy.read_input(os.path.join("tests", "inputs", "test_imp_redundant.imcnp"))


def test_default_importance_not_implemented():
    prob = montepy.read_input(os.path.join("tests", "inputs", "test_not_imp.imcnp"))
    prob.print_in_data_block["imp"] = True
    with pytest.raises(NotImplementedError):
        prob.write_problem(io.StringIO())



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import ABC, abstractmethod
import inspect
from functools import wraps


class SingletonGroup(ABC):
    """A base class for implementing a Singleton-like data structure.

    This treats immutable objects are Enums without having to list all.
    This is used for: Element, Nucleus, Library. When a brand new instance
    is requested it is created, cached and returned.
    If an existing instance is requested it is returned.
    This is done to reduce the memory usage for these objects.

    """

    __slots__ = "_instances"

    def __new__(cls, *args, **kwargs):
        kwargs_t = tuple([(k, v) for k, v in kwargs.items()])
        try:
            return cls._instances[args + kwargs_t]
        except KeyError:
            instance = super().__new__(cls)
            instance.__init__(*args, **kwargs)
            cls._instances[args + kwargs_t] = instance
            return cls._instances[args + kwargs_t]

    def __init_subclass__(cls, **kwargs):
        """Workaround to get sphinx autodoc happy."""
        cls._instances = {}
        super().__init_subclass__(**kwargs)

        original_new = cls.__new__

        @wraps(original_new)
        def __new__(cls, *args, **kwargs):
            return original_new(cls, *args, **kwargs)

        __new__.__signature__ = inspect.signature(cls.__init__)
        cls.__new__ = staticmethod(__new__)

    def __deepcopy__(self, memo):
        """Make deepcopy happy."""
        if self in memo:
            return memo[self]
        memo[self] = self
        return self

    @abstractmethod
    def __reduce__(self):
        """See: <https://docs.python.org/3/library/pickle.html#object.__reduce__>"""
        pass



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import io
import glob
import montepy
import os
import unittest
from tests import constants

import pickle

"""
This is for testing MontePy's interface with other libraries,
which are not dependencies.

This doesn't mean these interfaces are supported, but is meant to reduce user frustration
when using common workflows.
"""


class TestBrinyFermentation(unittest.TestCase):
    def test_pickling_all_models(self):
        for file in glob.glob(os.path.join("tests", "inputs", "*.imcnp")):
            if (
                os.path.basename(file) not in constants.BAD_INPUTS
                and os.path.basename(file) not in constants.IGNORE_FILES
            ):
                print(file)
                problem = montepy.read_input(file)
                with io.BytesIO() as fh:
                    pickle.dump(problem, fh)
                    fh.seek(0)
                    new_problem = pickle.load(fh)
                data = pickle.dumps(problem)
                new_problem = pickle.loads(data)
                for attr in {"cells", "surfaces", "materials", "data_inputs"}:
                    assert len(getattr(problem, attr)) == len(
                        getattr(new_problem, attr)
                    )
                    for obj in getattr(new_problem, attr):
                        assert obj._problem is not None
                        assert obj._problem is new_problem



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from __future__ import annotations
from abc import abstractmethod
import copy

import montepy
from montepy.errors import *
from montepy.input_parser.data_parser import (
    ClassifierParser,
    DataParser,
    ParamOnlyDataParser,
)
from montepy.input_parser.mcnp_input import Input
from montepy.particle import Particle
from montepy.mcnp_object import MCNP_Object, InitInput

import re
from typing import Union


class _ClassifierInput(Input):
    """A specialized subclass that returns only 1 useful token."""

    def tokenize(self):
        """Returns one token after all starting comments and spaces."""
        last_in_comment = True
        for token in super().tokenize():
            if token is None:
                break
            if last_in_comment:
                if token.type not in {"COMMENT", "SPACE"}:
                    last_in_comment = False
            else:
                if token.type == "SPACE":
                    break
            yield token


class DataInputAbstract(MCNP_Object):
    """Parent class to describe all MCNP data inputs.

    Parameters
    ----------
    input : Union[Input, str]
        the Input object representing this data input
    fast_parse : bool
        Whether or not to only parse the first word for the type of
        data.
    """

    _parser = DataParser()

    _classifier_parser = ClassifierParser()

    def __init__(
        self,
        input: InitInput = None,
        fast_parse=False,
    ):
        self._particles = None
        if not fast_parse:
            super().__init__(input, self._parser)
            if input:
                self.__split_name(input)
        else:
            if input:
                if isinstance(input, str):
                    input = _ClassifierInput(
                        input.split("\n"),
                        montepy.input_parser.block_type.BlockType.DATA,
                    )
                else:
                    input = copy.copy(input)
                    input.__class__ = _ClassifierInput
            super().__init__(input, self._classifier_parser)
            if input:
                self.__split_name(input)

    @staticmethod
    @abstractmethod
    def _class_prefix():
        """The text part of the input identifier.

        For example: for a material the prefix is ``m``

        this must be lower case

        Returns
        -------
        str
            the string of the prefix that identifies a input of this
            class.
        """
        pass

    @staticmethod
    @abstractmethod
    def _has_number():
        """Whether or not this class supports numbering.

        For example: ``kcode`` doesn't allow numbers but tallies do allow it e.g., ``f7``

        Returns
        -------
        bool
            True if this class allows numbers
        """
        pass

    @staticmethod
    @abstractmethod
    def _has_classifier():
        """Whether or not this class supports particle classifiers.

        For example: ``kcode`` doesn't allow particle types but tallies do allow it e.g., ``f7:n``

        * 0 : not allowed
        * 1 : is optional
        * 2 : is mandatory

        Returns
        -------
        int
            True if this class particle classifiers
        """
        pass

    @property
    def particle_classifiers(self):
        """The particle class part of the card identifier as a parsed list.

        This is parsed from the input that was read.

        For example: the classifier for ``F7:n`` is ``:n``, and ``imp:n,p`` is ``:n,p``
        This will be parsed as a list: ``[<Particle.NEUTRON: 'N'>, <Particle.PHOTON: 'P'>]``.

        Returns
        -------
        list
            the particles listed in the input if any. Otherwise None
        """
        if self._particles:
            return self._particles
        return None

    @property
    def prefix(self):
        """The text part of the card identifier parsed from the input.

        For example: for a material like: m20 the prefix is 'm'
        this will always be lower case.

        Returns
        -------
        str
            The prefix read from the input
        """
        return self._prefix.lower()

    @property
    def prefix_modifier(self):
        """The modifier to a name prefix that was parsed from the input.

        For example: for a transform: ``*tr5`` the modifier is ``*``

        Returns
        -------
        str
            the prefix modifier that was parsed if any. None if
            otherwise.
        """
        return self._modifier

    @property
    def data(self):
        """The syntax tree actually holding the data.

        Returns
        -------
        ListNode
            The syntax tree with the information.
        """
        return self._tree["data"]

    @property
    def classifier(self):
        """The syntax tree object holding the data classifier.

        For example this would container information like ``M4``, or ``F104:n``.

        Returns
        -------
        ClassifierNode
            the classifier for this data_input.
        """
        return self._tree["classifier"]

    def validate(self):
        pass

    def _update_values(self):
        pass

    def update_pointers(self, data_inputs):
        """Connects data inputs to each other

        Parameters
        ----------
        data_inputs : list
            a list of the data inputs in the problem

        Returns
        -------
        bool, None
            True iff this input should be removed from
            ``problem.data_inputs``
        """
        pass

    def __str__(self):
        return f"DATA INPUT: {self._tree['classifier']}"

    def __repr__(self):
        return str(self)

    def __split_name(self, input):
        """Parses the name of the data input as a prefix, number, and a particle classifier.

        This populates the properties:
            prefix
            _input_number
            classifier

        Parameters
        ----------
        input : input
            the input object representing this data input

        Raises
        ------
        MalformedInputError
            if the name is invalid for this DataInput
        """
        self._classifier = self._tree["classifier"]
        self.__enforce_name(input)
        self._input_number = self._classifier.number
        self._prefix = self._classifier._prefix.value
        if self._classifier.particles:
            self._particles = self._classifier.particles.particles
        self._modifier = self._classifier.modifier

    def __enforce_name(self, input):
        """Checks that the name is valid.

        Parameters
        ----------
        input : input
            the input object representing this data input

        Raises
        ------
        MalformedInputError
            if the name is invalid for this DataInput
        """
        classifier = self._classifier
        if self._class_prefix:
            if classifier.prefix.value.lower() != self._class_prefix():
                raise MalformedInputError(
                    input,
                    f"{self._tree['classifier'].format()} has the wrong prefix for {type(self)}",
                )
            if self._has_number():
                try:
                    num = classifier.number.value
                    assert num >= 0
                except (AttributeError, AssertionError) as e:
                    raise MalformedInputError(
                        input,
                        f"{classifier} does not contain a valid number",
                    )
            if not self._has_number() and classifier.number is not None:
                raise MalformedInputError(
                    input,
                    f"{classifier} cannot have a number for {type(self)}",
                )
            if self._has_classifier() == 2 and classifier.particles is None:
                raise MalformedInputError(
                    input,
                    f"{classifier} doesn't have a particle classifier for {type(self)}",
                )
            if self._has_classifier() == 0 and classifier.particles is not None:
                raise MalformedInputError(
                    input,
                    f"{classifier} cannot have a particle classifier for {type(self)}",
                )

    def __lt__(self, other):
        type_comp = self.prefix < other.prefix
        if type_comp:
            return type_comp
        elif self.prefix > other.prefix:
            return type_comp
        else:  # otherwise first part is equal
            return self._input_number.value < other._input_number.value


class DataInput(DataInputAbstract):
    """Catch-all for all other MCNP data inputs.

    Parameters
    ----------
    input : Union[Input, str]
        the Input object representing this data input
    fast_parse : bool
        Whether or not to only parse the first word for the type of
        data.
    prefix : str
        The input prefix found during parsing (internal use only)
    """

    def __init__(
        self, input: InitInput = None, fast_parse: bool = False, prefix: str = None
    ):
        if prefix:
            self._load_correct_parser(prefix)
        super().__init__(input, fast_parse)

    @property
    def _class_prefix(self):
        return None

    @property
    def _has_number(self):  # pragma: no cover
        return None

    @property
    def _has_classifier(self):  # pragma: no cover
        return None

    def _load_correct_parser(self, prefix):
        """Decides if a specialized parser needs to be loaded for barebone
        special cases.

        .. versionadded:: 0.3.0
        """
        PARAM_PARSER = ParamOnlyDataParser
        TALLY_PARSER = montepy.input_parser.tally_parser.TallyParser
        PARSER_PREFIX_MAP = {
            "f": TALLY_PARSER,
            "fm": TALLY_PARSER,
            "fs": montepy.input_parser.tally_seg_parser.TallySegmentParser,
            "sdef": PARAM_PARSER,
        }
        if prefix.lower() in PARSER_PREFIX_MAP:
            self._parser = PARSER_PREFIX_MAP[prefix.lower()]()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum, unique


@unique
class BlockType(Enum):
    """An enumeration for the different blocks in an input file."""

    CELL = 0
    """The first block that details Cell information."""
    SURFACE = 1
    """The second block that details Surface information."""
    DATA = 2
    """The third block that provides additional information
    and data.
    """



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.errors import *
from montepy.input_parser.parser_base import MCNP_Parser
from montepy.input_parser import syntax_node


class CellParser(MCNP_Parser):
    """The parser for parsing a Cell input.

    Returns
    -------
    SyntaxNode
        a syntax tree of the cell.
    """

    debugfile = None

    @_(
        "identifier_phrase material geometry_expr parameters",
        "identifier_phrase material geometry_expr",
        "padding identifier_phrase material geometry_expr parameters",
        "padding identifier_phrase material geometry_expr",
    )
    def cell(self, p):
        dict_tree = {}
        if isinstance(p[0], syntax_node.PaddingNode):
            dict_tree["start_pad"] = p[0]
        else:
            dict_tree["start_pad"] = syntax_node.PaddingNode()

        dict_tree["cell_num"] = p.identifier_phrase
        dict_tree["material"] = p.material
        dict_tree["geometry"] = p.geometry_expr
        if hasattr(p, "parameters"):
            dict_tree["parameters"] = p.parameters
        else:
            dict_tree["parameters"] = syntax_node.ParametersNode()
        return syntax_node.SyntaxNode("cell", dict_tree)

    @_("number_phrase KEYWORD")
    def cell(self, p):
        if p.KEYWORD.lower() == "like":
            raise UnsupportedFeature("The like but feature is not supported")

    @_(
        "null_ident_phrase",
        "identifier_phrase number_phrase",
        "identifier_phrase null_phrase",
    )
    def material(self, p):
        if len(p) == 1:
            ret_dict = {
                "mat_number": p.null_ident_phrase,
                "density": syntax_node.ValueNode(None, float),
            }
        else:
            ret_dict = {"mat_number": p[0], "density": p[1]}
        return syntax_node.SyntaxNode("material", ret_dict)

    @_('":"')
    def union(self, p):
        return syntax_node.PaddingNode(p[0])

    @_("union padding")
    def union(self, p):
        ret = p.union
        for node in p.padding.nodes:
            is_comment = isinstance(node, syntax_node.CommentNode)
            ret.append(node, is_comment)
        return ret

    @_("geometry_expr union geometry_term")
    def geometry_expr(self, p):
        left = p.geometry_expr
        right = p.geometry_term
        nodes = {"left": left, "operator": p.union, "right": right}
        return syntax_node.GeometryTree("union", nodes, ":", left, right)

    @_("geometry_term")
    def geometry_expr(self, p):
        term = p.geometry_term
        if isinstance(term, syntax_node.ValueNode):
            term = syntax_node.GeometryTree("shift", {"left": term}, ">", term)
        return term

    # handle intersection
    @_("geometry_term padding geometry_factor")
    def geometry_term(self, p):
        left = p.geometry_term
        right = p.geometry_factor
        nodes = {"left": left, "operator": p.padding, "right": right}
        return syntax_node.GeometryTree("intersection", nodes, "*", left, right)

    # handle implicit intersection of: ( )( )
    @_("geometry_term geometry_factory")
    def geometry_term(self, p):
        left = p.geometry_term
        right = p.geometry_factory
        nodes = {"left": left, "operator": syntax_node.PaddingNode(), "right": right}
        return syntax_node.GeometryTree("intersection", nodes, "*", left, right)

    @_(
        "geometry_term NUM_REPEAT",
        "geometry_term REPEAT",
        "geometry_term NUM_MULTIPLY",
        "geometry_term MULTIPLY",
        "geometry_term NUM_INTERPOLATE padding number_phrase",
        "geometry_term INTERPOLATE padding number_phrase",
        "geometry_term NUM_LOG_INTERPOLATE padding number_phrase",
        "geometry_term LOG_INTERPOLATE padding number_phrase",
    )
    def geometry_term(self, p):
        shortcut = syntax_node.ShortcutNode(p, data_type=int)
        node_iter = iter(shortcut.nodes)
        if isinstance(p.geometry_term, syntax_node.GeometryTree):
            left = p.geometry_term
            left.mark_last_leaf_shortcut(shortcut.type)
        else:
            left = next(node_iter)
        left_type = None
        if isinstance(left, syntax_node.ValueNode) and left == shortcut.nodes[0]:
            left_type = shortcut.type
        for node in node_iter:
            new_tree = syntax_node.GeometryTree(
                "intersection",
                {"left": left, "operator": syntax_node.PaddingNode(), "right": node},
                "*",
                left,
                node,
                left_short_type=left_type,
                right_short_type=shortcut.type,
            )
            left = new_tree
        return new_tree

    @_("geometry_term padding")
    def geometry_term(self, p):
        ret = p.geometry_term
        if isinstance(ret, syntax_node.ValueNode):
            if ret.padding:
                ret.padding.append(p.padding)
            else:
                ret.padding = p.padding
        else:
            if "end_pad" in ret.nodes:
                ret.nodes["end_pad"] += p.padding
            else:
                ret.nodes["end_pad"] = p.padding
        return ret

    @_("geometry_factor")
    def geometry_term(self, p):
        return p.geometry_factor

    @_("geometry_factory")
    def geometry_factor(self, p):
        return p.geometry_factory

    @_("COMPLEMENT geometry_factory")
    def geometry_factor(self, p):
        nodes = {
            "operator": syntax_node.PaddingNode(p.COMPLEMENT),
            "left": p.geometry_factory,
        }
        return syntax_node.GeometryTree(
            "complement",
            nodes,
            "#",
            p.geometry_factory,
        )

    @_("NUMBER")
    def geometry_factory(self, p):
        return syntax_node.ValueNode(p.NUMBER, float)

    @_('"(" geometry_expr ")"', '"(" padding geometry_expr ")"')
    def geometry_factory(self, p):
        nodes = {
            "start_pad": syntax_node.PaddingNode(p[0]),
            "left": p.geometry_expr,
            "end_pad": syntax_node.PaddingNode(p[-1]),
        }
        if hasattr(p, "padding"):
            for node in p.padding.nodes:
                nodes["start_pad"].append(node)
        return syntax_node.GeometryTree("geom parens", nodes, "()", p.geometry_expr)

    # support for fill card weirdness
    @_(
        'number_sequence "(" number_sequence ")"',
        'number_sequence "(" padding number_sequence ")"',
        'number_sequence "(" number_sequence ")" padding',
        'number_sequence "(" padding number_sequence ")" padding',
        'number_sequence ":" numerical_phrase',
        'number_sequence ":" padding numerical_phrase',
        # support for TRCL syntax
        '"(" number_sequence ")"',
        '"(" number_sequence ")" padding',
        '"(" padding number_sequence ")" padding',
    )
    def number_sequence(self, p):
        if isinstance(p[0], str):
            sequence = syntax_node.ListNode("parenthetical statement")
            sequence.append(p[0])
        else:
            sequence = p[0]
        for node in list(p)[1:]:
            if isinstance(node, syntax_node.ListNode):
                for val in node.nodes:
                    sequence.append(val)
            elif isinstance(node, str):
                sequence.append(syntax_node.PaddingNode(node))
            else:
                sequence.append(node)
        return sequence



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from io import StringIO
import pytest

import montepy
from montepy.errors import *
from montepy.input_parser import input_syntax_reader
from montepy.input_parser.mcnp_input import Input, Jump, Message, ReadInput, Title
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.input_file import MCNP_InputFile
from montepy.input_parser.parser_base import MCNP_Parser
from montepy.input_parser.shortcuts import Shortcuts
from montepy.input_parser import syntax_node
from montepy.particle import Particle
import warnings


class TestValueNode:
    def test_valuenoode_init(self):
        for type, token, answer in [
            (str, "hi", "hi"),
            (float, "1.2300", 1.23),
            (int, "1", 1),
            (float, "1.23e-3", 1.23e-3),
            (float, "6.02+23", 6.02e23),
        ]:
            for padding in [None, syntax_node.PaddingNode(" ")]:
                node = syntax_node.ValueNode(token, type, padding)
                assert node.value == answer
                assert node.token == token
                if padding:
                    assert node.padding == padding
                else:
                    assert node.padding is None
        # test with None values
        for type in {str, float, int}:
            node = syntax_node.ValueNode(None, type)
            assert None is (node.value)
            node = syntax_node.ValueNode(Jump(), type)
            assert None is (node.value)

    def test_valuenode_convert_to_int(self):
        node = syntax_node.ValueNode("1", float)
        node._convert_to_int()
        assert node.type == int
        assert node.value == 1
        # test 1.0
        node = syntax_node.ValueNode("1.0", float)
        node._convert_to_int()
        assert node.type == int
        assert node.value == 1
        # test wrong type
        with pytest.raises(ValueError):
            node = syntax_node.ValueNode("hi", str)
            node._convert_to_int()
        # test real float
        with pytest.raises(ValueError):
            node = syntax_node.ValueNode("1.23", float)
            node._convert_to_int()

    def test_valuenode_convert_to_enum(self):
        node = syntax_node.ValueNode("1", float)
        lat = montepy.data_inputs.lattice.Lattice
        node._convert_to_enum(lat)
        assert node.type == lat
        assert node.value == lat(1)
        # test with None
        with pytest.raises(ValueError):
            node = syntax_node.ValueNode(None, float)
            node._convert_to_enum(lat)
        node._convert_to_enum(lat, allow_none=True)
        assert None is (node.value)
        st = montepy.surfaces.surface_type.SurfaceType
        node = syntax_node.ValueNode("p", str)
        node._convert_to_enum(st, switch_to_upper=True)
        assert node.type == st
        assert node.value == st("P")

    def test_is_negat_identifier(self):
        node = syntax_node.ValueNode("-1", float)
        assert not node.is_negatable_identifier
        assert None is (node.is_negative)
        node.is_negatable_identifier = True
        assert node.is_negatable_identifier
        assert node.type == int
        assert node.value > 0
        assert node.is_negative
        # test with positive number
        node = syntax_node.ValueNode("1", float)
        node.is_negatable_identifier = True
        assert node.type == int
        assert node.value > 0
        assert not node.is_negative
        # test with none
        node = syntax_node.ValueNode(None, float)
        node.is_negatable_identifier = True
        assert node.type == int
        assert None is (node.value)
        assert None is (node.is_negative)
        node.value = 1
        assert node.value == 1
        assert not node.is_negative

    def test_is_negat_float(self):
        node = syntax_node.ValueNode("-1.23", float)
        assert not node.is_negatable_float
        assert None is (node.is_negative)
        node.is_negatable_float = True
        assert node.type == float
        assert node.value > 0
        assert node.is_negative
        assert node.is_negatable_float
        # test with positive number
        node = syntax_node.ValueNode("1.23", float)
        node.is_negatable_float = True
        assert node.type == float
        assert not node.is_negative
        # test with None
        node = syntax_node.ValueNode(None, float)
        node.is_negatable_float = True
        assert node.type == float
        assert None is (node.value)
        assert None is (node.is_negative)
        node.value = 1
        assert node.value == 1
        assert not node.is_negative

    def test_is_negative(self):
        node = syntax_node.ValueNode("-1.23", float)
        node.is_negatable_float = True
        assert node.is_negative
        node.is_negative = False
        assert node.value > 0
        assert not node.is_negative
        node = syntax_node.ValueNode("hi", str)
        node.is_negative = True
        assert None is (node.is_negative)

    @pytest.mark.parametrize(
        "token, val, answer, expand",
        [
            ("1", 5, "5", False),
            (1, 5, "5", False),
            ("-1", 2, " 2", False),
            ("-1", -2, "-2", False),
            ("+1", 5, "+5", False),
            ("0001", 5, "0005", False),
            (Jump(), 5, "5 ", False),
        ],
    )
    def test_valuenode_int_format(_, token, val, answer, expand):
        node = syntax_node.ValueNode("-1", int)
        output = node.format()
        assert output == "-1"
        node = syntax_node.ValueNode(token, int)
        node.value = val
        assert node.format() == answer
        node = syntax_node.ValueNode(token, int)
        node.is_negatable_identifier = True
        node.value = val
        node.is_negative = val < 0
        if expand:
            with pytest.warns(LineExpansionWarning):
                assert node.format() == answer
        else:
            assert node.format() == answer

    @pytest.mark.parametrize(
        "padding, val, answer, expand",
        [
            ([" "], 10, "10 ", True),
            (["  "], 10, "10 ", False),
            (["\n"], 10, "10\n", False),
            ([" ", "\n", "c hi"], 10, "10\nc hi", False),
            ([" ", "\n", "c hi"], 100, "100\nc hi", False),
            (["  ", "\n"], 10, "10 \n", False),
        ],
    )
    def test_value_node_format_padding(self, padding, val, answer, expand):
        pad_node = syntax_node.PaddingNode(padding[0])
        for pad in padding[1:]:
            pad_node.append(pad)
        node = syntax_node.ValueNode("1", int, pad_node)
        node.value = val
        if expand:
            warnings.simplefilter("default")
            with pytest.warns(LineExpansionWarning):
                assert node.format() == answer
        else:
            # change warnings to errors to ensure not raised
            warnings.resetwarnings()
            warnings.simplefilter("error")
            assert node.format() == answer

    def test_value_has_changed(self):
        # test None no change
        node = syntax_node.ValueNode(None, int)
        assert not node._value_changed
        # test None changed
        node.value = 5
        assert node._value_changed
        node = syntax_node.ValueNode("1.23", float)
        assert not node._value_changed
        node.value = 1.25
        assert node._value_changed
        node = syntax_node.ValueNode("hi", str)
        assert not node._value_changed
        node.value = "foo"
        assert node._value_changed

    def test_value_float_format(self):
        for input, val, answer, expand in [
            ("1.23", 1.23, "1.23", False),
            ("1.23", 4.56, "4.56", False),
            (1.23, 4.56, "4.56", False),
            ("-1.23", 4.56, " 4.56", False),
            ("1.0e-2", 2, "2.0e+0", False),
            ("1.602-19", 6.02e23, "6.020+23", False),
            ("1.602-0019", 6.02e23, "6.020+0023", False),
            (Jump(), 5.4, "5.4 ", True),
            ("1", 2, "2", False),
            ("0.5", 0, "0.0", False),
        ]:
            print("in:", input, "new value:", val, "answer:", answer)
            node = syntax_node.ValueNode(input, float)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with pytest.warns(LineExpansionWarning):
                    assert node.format() == answer
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                assert node.format() == answer

        for padding, val, answer, expand in [
            ([" "], 10, "10.0 ", True),
            (["  "], 10, "10.0 ", False),
            (["\n"], 10, "10.0\n", False),
            ([" ", "\n", "c hi"], 10, "10.0\nc hi", False),
            (["  ", "\n"], 10, "10.0 \n", False),
        ]:
            print("padding", padding, "new_val", val, "answer", answer)
            pad_node = syntax_node.PaddingNode(padding[0])
            for pad in padding[1:]:
                pad_node.append(pad)
            node = syntax_node.ValueNode("1.0", float, pad_node)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with pytest.warns(LineExpansionWarning):
                    assert node.format() == answer
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                assert node.format() == answer

    def test_value_str_format(self):
        for input, val, answer, expand in [
            ("hi", "foo", "foo", True),
            ("hi", None, "", False),
        ]:
            node = syntax_node.ValueNode(input, str)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with pytest.warns(LineExpansionWarning):
                    assert node.format() == answer
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                assert node.format() == answer
        for padding, val, answer, expand in [
            ([" "], "foo", "foo ", True),
            (["  "], "foo", "foo ", False),
            (["\n"], "foo", "foo\n", False),
            ([" ", "\n", "c hi"], "foo", "foo\nc hi", False),
            (["  ", "\n"], "foo", "foo \n", False),
        ]:
            print("padding", padding, "new_val", val, "answer", answer)
            pad_node = syntax_node.PaddingNode(padding[0])
            for pad in padding[1:]:
                pad_node.append(pad)
            node = syntax_node.ValueNode("hi", str, pad_node)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with pytest.warns(LineExpansionWarning):
                    assert node.format() == answer
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                assert node.format() == answer

    def test_value_enum_format(self):
        lat = montepy.data_inputs.lattice.Lattice
        st = montepy.surfaces.surface_type.SurfaceType
        for input, val, enum_class, args, answer, expand in [
            (
                "1",
                lat.HEXAGONAL,
                lat,
                {"format_type": int, "switch_to_upper": False},
                "2",
                False,
            ),
            (
                1,
                lat.HEXAGONAL,
                lat,
                {"format_type": int, "switch_to_upper": False},
                "2",
                False,
            ),
            ("p", st.PZ, st, {"format_type": str, "switch_to_upper": True}, "PZ", True),
        ]:
            node = syntax_node.ValueNode(input, args["format_type"])
            node._convert_to_enum(enum_class, **args)
            node.value = val
            if expand:
                warnings.simplefilter("default")
                with pytest.warns(LineExpansionWarning):
                    assert node.format() == answer
            else:
                # change warnings to errors to ensure not raised
                warnings.resetwarnings()
                warnings.simplefilter("error")
                assert node.format() == answer

    def test_value_comments(self):
        value_node = syntax_node.ValueNode("1", int)
        assert len(list(value_node.comments)) == 0
        padding = syntax_node.PaddingNode("$ hi", True)
        value_node.padding = padding
        comments = list(value_node.comments)
        assert len(comments) == 1
        assert "hi" in comments[0].contents

    def test_value_trailing_comments(self):
        value_node = syntax_node.ValueNode("1", int)
        assert None is (value_node.get_trailing_comment())
        value_node._delete_trailing_comment()
        assert None is (value_node.get_trailing_comment())
        padding = syntax_node.PaddingNode("c hi", True)
        value_node.padding = padding
        comment = value_node.get_trailing_comment()
        assert len(comment) == 1
        assert comment[0].contents == "hi"
        value_node._delete_trailing_comment()
        assert None is (value_node.get_trailing_comment())

    def test_value_str(self):
        value_node = syntax_node.ValueNode("1", int)
        str(value_node)
        repr(value_node)
        padding = syntax_node.PaddingNode("$ hi", True)
        value_node.padding = padding
        str(value_node)
        repr(value_node)

    def test_value_equality(self):
        value_node1 = syntax_node.ValueNode("1", int)
        assert value_node1 == value_node1
        assert not value_node1 == syntax_node.PaddingNode("")
        value_node2 = syntax_node.ValueNode("2", int)
        assert value_node1 != value_node2
        value_node3 = syntax_node.ValueNode("hi", str)
        assert value_node1 != value_node3
        assert value_node1 == 1
        assert value_node1 != 2
        assert value_node1 != "hi"
        value_node4 = syntax_node.ValueNode("1.5", float)
        value_node5 = syntax_node.ValueNode("1.50000000000001", float)
        assert value_node4 == value_node5
        value_node5.value = 2.0
        assert value_node4 != value_node5


class TestSyntaxNode:
    @pytest.fixture
    def test_node(_):
        value1 = syntax_node.ValueNode("1.5", float)
        value2 = syntax_node.ValueNode("1", int)
        test_node = syntax_node.SyntaxNode(
            "test",
            {"foo": value1, "bar": value2, "bar2": syntax_node.SyntaxNode("test2", {})},
        )
        return test_node

    def test_syntax_init(_, test_node):
        test = test_node
        assert test.name == "test"
        assert "foo" in test.nodes
        assert "bar" in test.nodes
        assert isinstance(test.nodes["foo"], syntax_node.ValueNode)

    def test_syntax_name(_, test_node):
        test = test_node
        test.name = "hi"
        assert test.name == "hi"
        with pytest.raises(TypeError):
            test.name = 1.0

    def test_get_value(_, test_node):
        test = test_node
        assert test.get_value("foo") == 1.5
        with pytest.raises(KeyError):
            test.get_value("foo2")
        with pytest.raises(KeyError):
            test.get_value("bar2")

    def test_syntax_format(_, test_node):
        output = test_node.format()
        assert output == "1.51"

    def test_syntax_dict(_, test_node):
        test = test_node
        assert "foo" in test
        assert test["foo"] == test.nodes["foo"]

    def test_syntax_comments(_, test_node):
        padding = syntax_node.PaddingNode("$ hi", True)
        test = copy.deepcopy(test_node)
        test["foo"].padding = padding
        padding = syntax_node.PaddingNode("$ foo", True)
        test["bar"].padding = padding
        comments = list(test.comments)
        assert len(comments) == 2

    def test_syntax_trailing_comments(_, test_node):
        # test with blank tail
        assert None is test_node.get_trailing_comment()
        test = copy.deepcopy(test_node)
        test["bar2"].nodes["foo"] = syntax_node.ValueNode("1.23", float)
        assert None is (test.get_trailing_comment())
        test["bar2"]["foo"].padding = syntax_node.PaddingNode("c hi", True)
        assert len(test.get_trailing_comment()) == 1
        test._delete_trailing_comment()
        assert None is (test.get_trailing_comment())

    def test_syntax_str(_, test_node):
        str(test_node)
        repr(test_node)
        test_node._pretty_str()


class TestGeometryTree:

    @pytest.fixture
    def test_tree(_):
        left = syntax_node.ValueNode("1", int)
        right = syntax_node.ValueNode("2", int)
        op = syntax_node.PaddingNode(" ")
        return syntax_node.GeometryTree(
            "test",
            {"left": left, "operator": op, "right": right},
            montepy.Operator.INTERSECTION,
            left,
            right,
        )

    def test_geometry_init(self):
        left = syntax_node.ValueNode("1", int)
        right = syntax_node.ValueNode("2", int)
        op = syntax_node.PaddingNode(" ")
        tree = syntax_node.GeometryTree(
            "test",
            {"left": left, "operator": op, "right": right},
            montepy.Operator.INTERSECTION,
            left,
            right,
        )
        assert tree.left is left
        assert tree.right is right
        assert tree.operator == montepy.Operator.INTERSECTION

    def test_geometry_format(_, test_tree):
        assert test_tree.format() == "1 2"

    def test_geometry_str(_, test_tree):
        test = test_tree
        str(test)
        repr(test)
        test._pretty_str()

    def test_geometry_comments(_, test_tree):
        test = copy.deepcopy(test_tree)
        test.left.padding = syntax_node.PaddingNode("$ hi", True)
        comments = list(test.comments)
        assert len(comments) == 1


def test_geometry_tree_mark_last_leaf_shortcut():
    # test left leaf
    tree = {
        "left": syntax_node.ValueNode("123", int),
        "operator": syntax_node.PaddingNode(" "),
    }
    geom = syntax_node.GeometryTree(
        "test", tree, montepy.Operator.INTERSECTION, tree["left"]
    )
    geom.mark_last_leaf_shortcut(Shortcuts.REPEAT)
    assert geom._left_short_type == Shortcuts.REPEAT
    # try overriding old shortcut
    geom.mark_last_leaf_shortcut(Shortcuts.MULTIPLY)
    assert geom._left_short_type == Shortcuts.REPEAT
    tree["right"] = syntax_node.ValueNode("789", int)
    geom = syntax_node.GeometryTree(
        "test", tree, montepy.Operator.INTERSECTION, tree["left"], tree["right"]
    )
    del tree["right"]
    tree["left"] = syntax_node.ValueNode("456", int)
    # make unbalanced tree in the other way for rigor
    geom = syntax_node.GeometryTree(
        "test", tree, montepy.Operator.INTERSECTION, tree["left"], geom
    )
    # test a right-side leaf
    geom.mark_last_leaf_shortcut(Shortcuts.REPEAT)
    assert geom.right._right_short_type == Shortcuts.REPEAT
    # try overriding old shortcut
    geom.mark_last_leaf_shortcut(Shortcuts.MULTIPLY)
    assert geom.right._right_short_type == Shortcuts.REPEAT


class TestPaddingNode:
    def test_padding_init(self):
        pad = syntax_node.PaddingNode(" ")
        assert len(pad.nodes) == 1
        assert pad.value == " "

    def test_padding_is_space(self):
        pad = syntax_node.PaddingNode(" ")
        assert pad.is_space(0)
        pad.append("\n")
        assert not pad.is_space(1)
        pad.append("$ hi", True)
        assert not pad.is_space(2)
        with pytest.raises(IndexError):
            pad.is_space(5)

    def test_padding_append(self):
        pad = syntax_node.PaddingNode(" ")
        pad.append("\n")
        assert len(pad) == 2
        pad.append(" ")
        assert len(pad) == 3
        pad.append(" \n")
        assert len(pad) == 5
        pad.append("$ hi", True)
        assert len(pad) == 6

    def test_padding_format(self):
        pad = syntax_node.PaddingNode(" ")
        assert pad.format() == " "
        pad.append("$ hi", True)
        assert pad.format() == " $ hi"

    def test_padding_grab_beginning_format(self):
        pad = syntax_node.PaddingNode(" ")
        new_pad = [
            syntax_node.CommentNode("c hi"),
            "\n",
            syntax_node.CommentNode("c foo"),
        ]
        answer = copy.copy(new_pad)
        pad._grab_beginning_comment(new_pad)
        assert pad.nodes == answer + ["\n", " "]

    def test_padding_eq(self):
        pad = syntax_node.PaddingNode(" ")
        assert pad == " "
        assert pad != " hi "
        pad1 = syntax_node.PaddingNode(" ")
        assert pad == pad1
        assert not pad == 1

    def test_comment_init(self):
        comment = syntax_node.CommentNode("$ hi")
        assert isinstance(comment.nodes[0], syntax_node.SyntaxNode)
        assert len(comment.nodes) == 1
        assert comment.is_dollar
        comment = syntax_node.CommentNode(" c hi")
        assert not comment.is_dollar
        assert len(list(comment.comments)) == 1

    def test_comment_append(self):
        comment = syntax_node.CommentNode("c foo")
        comment.append("c bar")
        assert len(comment.nodes) == 2
        # test mismatch
        comment = syntax_node.CommentNode("$ hi")
        with pytest.raises(TypeError):
            comment.append("c hi")

    def test_comment_str(self):
        comment = syntax_node.CommentNode("$ hi")
        str(comment)
        repr(comment)

    def test_blank_dollar_comment(self):
        comment = syntax_node.CommentNode("$")
        assert comment.is_dollar
        assert len(list(comment.comments)) == 1
        assert len(comment.contents) == 0


def test_graveyard_comment():
    padding = syntax_node.PaddingNode(" ")
    padding.append("$ test", True)
    assert padding.has_graveyard_comment()
    padding.append(" ")
    assert padding.has_graveyard_comment()
    padding.append("\n")
    assert not padding.has_graveyard_comment()


@pytest.mark.parametrize(
    "padding,expect",
    [
        (["$ hi"], False),
        ([" c style comment"], False),
        ([" "], True),
        (["$ hi", "    ", "c hi"], True),
    ],
)
def test_padding_has_space(padding, expect):
    node = syntax_node.PaddingNode(padding[0])
    for pad in padding[1:]:
        node.append(pad)
    assert node.has_space() == expect


class TestParticlesNode:
    def test_particle_init(self):
        parts = syntax_node.ParticleNode("test", ":n,p,e")
        particle = montepy.particle.Particle
        answers = {particle.NEUTRON, particle.PHOTON, particle.ELECTRON}
        assert parts.particles == answers
        assert len(list(parts.comments)) == 0
        for part in parts:
            assert part in answers
        answers = [particle.NEUTRON, particle.PHOTON, particle.ELECTRON]
        assert parts._particles_sorted == answers

    def test_particles_setter(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        particle = montepy.particle.Particle
        parts.particles = {particle.TRITON}
        assert parts.particles == {particle.TRITON}
        parts.particles = [particle.TRITON]
        assert parts.particles == {particle.TRITON}
        with pytest.raises(TypeError):
            parts.particles = "hi"
        with pytest.raises(TypeError):
            parts.particles = {"hi"}

    def test_particles_add_remove(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        particle = montepy.particle.Particle
        parts.add(particle.TRITON)
        assert particle.TRITON in parts
        assert parts._particles_sorted[-1] == particle.TRITON
        with pytest.raises(TypeError):
            parts.add("hi")
        parts.remove(particle.NEUTRON)
        assert particle.NEUTRON not in parts
        with pytest.raises(TypeError):
            parts.remove("hi")

    def test_particles_sorted(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        particle = montepy.particle.Particle
        # lazily work around internals
        parts._particles.remove(particle.NEUTRON)
        assert particle.NEUTRON not in parts._particles_sorted
        parts._particles.add(particle.TRITON)
        assert particle.TRITON in parts._particles_sorted

    def test_particles_format(self):
        parts = syntax_node.ParticleNode("test", "n,p,e")
        repr(parts)
        assert parts.format() == ":n,p,e"
        parts = syntax_node.ParticleNode("test", "N,P,E")
        assert parts.format() == ":N,P,E"


class TestListNode:
    def test_list_init(self):
        list_node = syntax_node.ListNode("list")
        assert list_node.nodes == []

    def test_list_append(self):
        list_node = syntax_node.ListNode("list")
        list_node.append(syntax_node.ValueNode("1.0", float))
        assert len(list_node) == 1

    def test_list_str(self):
        list_node = syntax_node.ListNode("list")
        list_node.append(syntax_node.ValueNode("1.0", float))
        str(list_node)
        repr(list_node)
        list_node._pretty_str()

    def test_list_slicing(self):
        list_node = syntax_node.ListNode("list")
        for i in range(20):
            list_node.append(syntax_node.ValueNode("1.0", float))
        assert list_node[5] == syntax_node.ValueNode("1.0", float)
        for val in list_node[1:5]:
            assert val == syntax_node.ValueNode("1.0", float)
        for val in list_node[1:5:1]:
            assert val == syntax_node.ValueNode("1.0", float)
        for val in list_node[::1]:
            assert val == syntax_node.ValueNode("1.0", float)
        for val in list_node[5:1:-1]:
            assert val == syntax_node.ValueNode("1.0", float)
        for val in list_node[::-1]:
            assert val == syntax_node.ValueNode("1.0", float)
        with pytest.raises(IndexError):
            list_node[50]

    def test_list_equality(self):
        list_node1 = syntax_node.ListNode("list")
        for i in range(20):
            list_node1.append(syntax_node.ValueNode("1.0", float))
        assert not list_node1 == "hi"
        list2 = [syntax_node.ValueNode("1.0", float)] * 19
        assert not list_node1 == list2
        list2 = [syntax_node.ValueNode("1.0", float)] * 20
        assert list_node1 == list2
        list2 = [syntax_node.ValueNode("1.0", float)] * 19 + [
            syntax_node.ValueNode("1.5", float)
        ]
        assert list_node1 != list2

    def test_list_trailing_comment(self):
        list_node1 = syntax_node.ListNode("list")
        for i in range(20):
            list_node1.append(syntax_node.ValueNode("1.0", float))
        padding = syntax_node.PaddingNode("c hi", True)
        list_node1[-1].padding = padding
        comments = list(list_node1.get_trailing_comment())
        assert len(comments) == 1
        list_node1._delete_trailing_comment()
        assert None is (list_node1.get_trailing_comment())
        # test an empty list
        list_node1 = syntax_node.ListNode("list")
        assert None is (list_node1.get_trailing_comment())
        list_node1._delete_trailing_comment()
        assert None is (list_node1.get_trailing_comment())

    def test_list_format(self):
        list_node = syntax_node.ListNode("list")
        for i in range(20):
            list_node.append(syntax_node.ValueNode("1.0", float))
        assert list_node.format() == "1.0 " * 19 + "1.0"

    def test_list_comments(self):
        list_node = syntax_node.ListNode("list")
        for i in range(20):
            list_node.append(syntax_node.ValueNode("1.0", float))
        padding = syntax_node.PaddingNode("$ hi", True)
        list_node[-1].padding = padding
        comments = list(list_node.comments)
        assert len(comments) == 1


class TestMaterialssNode:
    def test_isotopes_init(self):
        isotope = syntax_node.MaterialsNode("test")
        assert isotope.name == "test"
        assert isinstance(isotope.nodes, list)

    def test_isotopes_append(self):
        isotopes = syntax_node.MaterialsNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        concentration = syntax_node.ValueNode("1.5", float)
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        assert isotopes.nodes[-1][0] == zaid
        assert isotopes.nodes[-1][1] == concentration

    def test_isotopes_format(self):
        padding = syntax_node.PaddingNode(" ")
        isotopes = syntax_node.MaterialsNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        zaid.padding = padding
        concentration = syntax_node.ValueNode("1.5", float)
        concentration.padding = padding
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        assert isotopes.format() == "1001.80c 1.5 "

    def test_isotopes_str(self):
        isotopes = syntax_node.MaterialsNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        concentration = syntax_node.ValueNode("1.5", float)
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        str(isotopes)
        repr(isotopes)
        isotopes._pretty_str()

    def test_isotopes_iter(self):
        isotopes = syntax_node.MaterialsNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        concentration = syntax_node.ValueNode("1.5", float)
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        for combo in isotopes:
            assert len(combo) == 2

    def test_isotopes_comments(self):
        padding = syntax_node.PaddingNode(" ")
        isotopes = syntax_node.MaterialsNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        zaid.padding = padding
        concentration = syntax_node.ValueNode("1.5", float)
        padding = copy.deepcopy(padding)
        padding.append("$ hi", True)
        concentration.padding = padding
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        comments = list(isotopes.comments)
        assert len(comments) == 1
        assert comments[0].contents == "hi"

    def test_isotopes_trailing_comment(self):
        padding = syntax_node.PaddingNode(" ")
        isotopes = syntax_node.MaterialsNode("test")
        zaid = syntax_node.ValueNode("1001.80c", str)
        zaid.padding = padding
        concentration = syntax_node.ValueNode("1.5", float)
        padding = copy.deepcopy(padding)
        padding.append("c hi", True)
        concentration.padding = padding
        isotopes.append_nuclide(("isotope_fraction", zaid, concentration))
        comments = isotopes.get_trailing_comment()
        assert len(comments) == 1
        assert comments[0].contents == "hi"
        isotopes._delete_trailing_comment()
        comments = isotopes.get_trailing_comment()
        assert None is (comments)


class TestShortcutNode:
    def test_basic_shortcut_init(self):
        with pytest.raises(ValueError):
            syntax_node.ShortcutNode("")
        # test a blank init
        shortcut = syntax_node.ShortcutNode(
            short_type=syntax_node.Shortcuts.LOG_INTERPOLATE
        )
        assert shortcut._type == syntax_node.Shortcuts.LOG_INTERPOLATE
        assert shortcut.end_padding.nodes == [" "]
        with pytest.raises(TypeError):
            syntax_node.ShortcutNode(short_type="")

    def test_shortcut_end_padding_setter(self):
        short = syntax_node.ShortcutNode()
        pad = syntax_node.PaddingNode(" ")
        short.end_padding = pad
        assert short.end_padding == pad
        with pytest.raises(TypeError):
            short.end_padding = " "


"""
Most examples, unless otherwise noted are taken from Section 2.8.1
of LA-UR-17-29981.
"""
tests = {
    "1 3M 2r": [1, 3, 3, 3],
    # unofficial
    "0.01 2ILOG 10": [0.01, 0.1, 1, 10],
    "1 3M I 4": [1, 3, 3.5, 4],
    "1 3M 3M": [1, 3, 9],
    "1 2R 2I 2.5": [1, 1, 1, 1.5, 2, 2.5],
    "1 R 2m": [1, 1, 2],
    "1 R R": [1, 1, 1],
    "1 2i 4 3m": [1, 2, 3, 4, 12],
    # unofficial
    "1 i 3": [1, 2, 3],
    # unofficial
    "1 ilog 100": [1, 10, 100],
    "1 1r ilog 100": [1, 1, 10, 100],
    # last official one
    "1 2i 4 2i 10": [
        1,
        2,
        3,
        4,
        6,
        8,
        10,
    ],
    "1 2j 4": [1, montepy.Jump(), montepy.Jump(), 4],
    "1 j": [1, montepy.Jump()],
}


@pytest.mark.parametrize("test, answer", tests.items())
def test_shortcut_expansion_valid(test, answer):
    parser = ShortcutTestFixture()
    print(test)
    input = Input([test], BlockType.DATA)
    parsed = parser.parse(input.tokenize())
    print(parsed)
    for val, gold in zip(parsed, answer):
        if val.value is None:
            assert gold == montepy.Jump()
        else:
            assert val.value == pytest.approx(gold)
    assert parsed.format() == test


@pytest.mark.parametrize(
    "test",
    [
        ("3J 4R"),
        ("1 4I 3M"),
        # last official test
        ("1 4I J"),
        ("1 2Ilog J"),
        ("J 2Ilog 5"),
        ("3J 2M"),
        ("10 M"),
        ("2R"),
    ],
)
def test_shortcut_expansion_invalid(test):
    print(test)
    parser = ShortcutTestFixture()
    with pytest.raises(ValueError):
        input = Input([test], BlockType.DATA)
        parsed = parser.parse(input.tokenize())
        if parsed is None:
            raise montepy.errors.MalformedInputError("", "")


@pytest.mark.parametrize(
    "test, answer, form_ans",
    [
        ("1 3r ", [1, 1, 1, 1], None),
        ("1 1 3r ", [1, 1, 1, 1, 1], None),
        ("1 1 2M 3r ", [1, 1, 2, 2, 2, 2], None),
        ("1 -2M ", [1, -2], None),
        ("1 2i 4 ", [1, 2, 3, 4], None),
        ("1 2i 4 ", [1, 2, 3, 4], None),
        ("1 1 2i 4 ", [1, 1, 2, 3, 4], None),
        ("1 1 2i 4 5 6 ", [1, 1, 2, 3, 4, 5, 6], "1 1 2I 4  5 6"),
        ("1 1 2i 4:5 6 ", [1, 1, 2, 3, 4, 5, 6], "1 1 2I 4 :5 6"),
        ("1 ilog 100 ", [1, 10, 100], "1 1ILOG 100"),
        ("1 1r ilog 100 ", [1, 1, 10, 100], "1 1R 1ILOG 100"),
        # secretly test iterator
        ("#1", [1], None),
        ("#(1 2 3)", [1, 2, 3], None),
        ("1 2:( 3 4 5)", [1, 2, 3, 4, 5], None),
    ],
)
def test_shortcut_geometry_expansion(test, answer, form_ans):

    parser = ShortcutGeometryTestFixture()
    print(test)
    input = Input([test], BlockType.CELL)
    parsed = parser.parse(input.tokenize())
    for val, gold in zip(parsed, answer):
        assert val.value == gold
    if form_ans:
        assert parsed.format().rstrip() == form_ans
    else:
        assert parsed.format().rstrip() == test.upper().rstrip()


@pytest.mark.parametrize(
    "test, length, indices",
    [
        ("1 3r ", 1, {0: Shortcuts.REPEAT}),
        ("1 1 3r ", 2, {1: Shortcuts.REPEAT}),
        ("1 1 2M 3r ", 3, {1: Shortcuts.MULTIPLY, 2: Shortcuts.REPEAT}),
        ("1 -2M ", 1, {0: Shortcuts.MULTIPLY}),
        ("1 2i 4 ", 1, {0: Shortcuts.INTERPOLATE}),
        ("1 2i 4 ", 1, {0: Shortcuts.INTERPOLATE}),
        ("1 1 2i 4 ", 2, {1: Shortcuts.INTERPOLATE}),
        ("1 1 2i 4 5 6 ", 4, {1: Shortcuts.INTERPOLATE}),
        ("1 1 2i 4:5 6 ", 4, {1: Shortcuts.INTERPOLATE}),
        ("1 ilog 100 ", 1, {0: Shortcuts.LOG_INTERPOLATE}),
        # secretly test iterator
        ("#1", 1, {}),
        ("#(1 2 3)", 3, {}),
        ("1 2:( 3 4 5)", 5, {}),
    ],
)
def test_shortcut_flatten(test, length, indices):

    parser = ShortcutGeometryTestFixture()
    print(test)
    input = Input([test], BlockType.CELL)
    parsed = parser.parse(input.tokenize())
    flatpack = parsed._flatten_shortcut()
    assert len(flatpack) == length
    for index, short_type in indices.items():
        assert isinstance(flatpack.nodes[index], syntax_node.ShortcutNode)
        assert flatpack.nodes[index].type == short_type


@pytest.mark.parametrize(
    "in_str, answer",
    [
        ("1 5R", "1 5R"),
        ("1 5r", "1 5r"),
        ("1 5r 2m", "1 5r 2m"),
        ("1 r", "1 r"),
        ("1 r 2 2r", "1 r 2 2r"),
        ("1 J 5 2R", "1 J 5 2R"),
        ("1 2i 5", "1 2i 5"),
        ("1 2ilog 5", "1 2ilog 5"),
        ("1 r 2i 5", "1 r 2i 5"),
        ("1 2i 5 r", "1 2i 5 r"),
        ("1 r 2ilog 5", "1 r 2ilog 5"),
        ("1 5M", "1 5M"),
        ("1 5m", "1 5m"),
        ("1 5m 2r", "1 5m 2r"),
        ("2j ", "2j "),
        ("2j", "2j"),
        ("2J ", "2J "),
        ("J", "J"),
        ("0 2i 3 10 2i 16 1", "0 2i 3 10 2i 16 1"),
    ],
)
def test_shortcut_format(in_str, answer):
    parser = ShortcutTestFixture()
    input = Input([in_str], BlockType.CELL)
    shortcut = parser.parse(input.tokenize())
    assert shortcut.format() == answer
    # try jump with empty jump shortcut
    shortcut.nodes.clear()
    assert shortcut.format() == ""


@pytest.mark.parametrize(
    "testee, outcome",
    [("(", False), (1.0, True), (0.0, True), ("s", False), (5.0, False)],
)
def test_shortcut_contains(testee, outcome):
    parser = ShortcutTestFixture()
    input = Input(["1 0 2R"], BlockType.CELL)
    node = parser.parse(input.tokenize())
    assert (testee in node) == outcome, "Contains result was incorrect."


class ShortcutTestFixture(MCNP_Parser):
    @_("number_sequence", "shortcut_sequence")
    def shortcut_magic(self, p):
        return p[0]


class ShortcutGeometryTestFixture(montepy.input_parser.cell_parser.CellParser):
    @_("geometry_expr")
    def geometry(self, p):
        return p[0]


class TestShortcutListIntegration:
    @pytest.fixture
    def parser(_):
        return ShortcutTestFixture()

    @pytest.fixture
    def list_node(_, parser):
        input = Input(["1 1 2i 4"], BlockType.DATA)
        return parser.parse(input.tokenize())

    def test_shortcut_list_node_init(_, list_node):
        answers = [1, 1, 2, 3, 4]
        for val, gold in zip(list_node, answers):
            assert val.value == pytest.approx(gold)

    def test_shortcut_list_update_vals(_, list_node):
        list_node = copy.deepcopy(list_node)
        values = list(list_node)
        list_node.update_with_new_values(values)
        assert list(list_node) == values

    def test_shortcut_list_update_vals_repeat(_, parser):
        input = Input(["1 2 3 5R 0 0"], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.insert(2, syntax_node.ValueNode(3.0, float))
        list_node.update_with_new_values(values)
        assert len(list_node._shortcuts[0].nodes) == 7
        assert list(list_node) == values

    def test_shortcut_list_trailing_jump(_, parser):
        input = Input(["1 2 3 5R 0 0"], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode(Jump(), float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values[:-1]
        # test with User specified end jump
        input = Input(["1 2 3 5R 0 0 j"], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode(Jump(), float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values

    def test_shortcut_list_touching_shortcuts(_, parser):
        input = Input(["1 2 3 5R 3 3 4R 0 0"], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        list_node.update_with_new_values(values)
        assert list(list_node) == values

    def test_shortcut_then_jump_compress(_, parser):
        for input_str, index in [
            ("1 2 3 5R 3 3 4R 0 0", -2),
            ("1 2 3M 4", -2),
            ("1 2ilog 10", -2),
        ]:
            print(input_str)
            input = Input([input_str], BlockType.DATA)
            list_node = parser.parse(input.tokenize())
            values = list(list_node)
            values[index].value = None
            list_node.update_with_new_values(values)
            assert list(list_node) == values

    def test_shortcut_list_shortcut_cant_consume(_, parser):
        # try with wrong type
        input = Input(["1 2 3 5R 3 3 4R "], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("hi", str))
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[1].nodes) == 5
        # try with wrong value
        input = Input(["1 2 3 5R 3 3 4R "], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("5.0", float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[1].nodes) == 5
        # try with right value
        values[-1].value = 3.0
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[1].nodes) == 6

    def test_shortcut_list_multiply(_, parser):
        input = Input(["1 2 5M "], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("5.0", float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[0].nodes) == 2
        input = Input(["0.5 2R 5M "], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("5.0", float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[1].nodes) == 1

    def test_shortcut_list_interpolate(_, parser):
        # try with log interpolate
        input = Input(["1.0 0.01 2ILOG 10"], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("100", float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[0].nodes) == 5
        # try with linear interpolate
        input = Input(["1 1 2I 2.5"], BlockType.DATA)
        list_node = parser.parse(input.tokenize())
        values = list(list_node)
        values.append(syntax_node.ValueNode("3", float))
        list_node.update_with_new_values(values)
        assert list(list_node) == values
        assert len(list_node._shortcuts[0].nodes) == 5


class TestSyntaxParsing:
    def testCardInit(self):
        with pytest.raises(TypeError):
            Input("5", BlockType.CELL)
        with pytest.raises(TypeError):
            Input([5], BlockType.CELL)
        with pytest.raises(TypeError):
            Input(["5"], "5")

    def testMessageInit(self):
        with pytest.raises(TypeError):
            Message(["hi"], "5")
        with pytest.raises(TypeError):
            Message(["hi"], [5])

    def testTitleInit(self):
        with pytest.raises(TypeError):
            Title(["hi"], 5)

    def testMessageFinder(self):
        test_message = "this is a message"
        test_string = f"""message: {test_message}

test title
"""
        for tester, validator in [
            (test_string, test_message),
            (test_string.upper(), test_message.upper()),
        ]:
            with StringIO(tester) as fh:
                generator = input_syntax_reader.read_front_matters(fh, (6, 2, 0))
                card = next(generator)
                assert isinstance(card, montepy.input_parser.mcnp_input.Message)
                assert card.lines[0] == validator
                assert len(card.lines) == 1

    def testReadCardStr(self):
        card = ReadInput(["Read file=hi.imcnp"], BlockType.CELL)
        assert str(card) == "READ INPUT: Block_Type: BlockType.CELL"
        assert (
            repr(card)
            == "READ INPUT: BlockType.CELL: ['Read file=hi.imcnp'] File: hi.imcnp"
        )

    def testSingleLineEndComment(self):
        # tests issues #117
        input = Input(["c"], BlockType.CELL)
        generator = input.tokenize()
        token = next(generator)
        assert token.type == "COMMENT"
        assert token.value == "c"

    def testReadCardConfusions(self):
        for file in {"A1_cells", "I1_cells"}:
            input = ReadInput([f"Read FILE={file}"], BlockType.CELL)
            assert input.file_name == file

    def testReadCardBadSyntax(self):
        with pytest.raises(ParsingError):
            card = ReadInput(["Read 1"], BlockType.CELL)

    def testTitleFinder(self):
        test_title = "Richard Stallman writes GNU"
        test_string = f"""{test_title}
1 0 -1
"""
        for tester, validator in [
            (test_string, test_title),
            (test_string.upper(), test_title.upper()),
        ]:
            with StringIO(tester) as fh:
                generator = input_syntax_reader.read_front_matters(fh, (6, 2, 0))
                card = next(generator)
                assert isinstance(card, montepy.input_parser.mcnp_input.Title)
                assert card.title == validator

    def testCardFinder(self):
        test_string = """1 0 -1
     5"""
        for i in range(5):
            tester = " " * i + test_string
            with StringIO(tester) as fh:
                fh.lineno = 0
                fh.path = "foo"
                generator = input_syntax_reader.read_data(fh, (6, 2, 0))
                card = next(generator)
                assert isinstance(card, montepy.input_parser.mcnp_input.Input)
                answer = [" " * i + "1 0 -1", "     5"]
                assert len(answer) == len(card.input_lines)
                for j, line in enumerate(card.input_lines):
                    assert line == answer[j]
                assert card.block_type == montepy.input_parser.block_type.BlockType.CELL

    def testReadCardFinder(self):
        test_string = "read file=foo.imcnp "
        with StringIO(test_string) as fh:
            fh.lineno = 0
            fh.path = "foo"
            card = next(input_syntax_reader.read_data(fh, (6, 2, 0)))
            assert None is (card)  # the read input is hidden from the user

    def testBlockId(self):
        test_string = "1 0 -1"
        for i in range(3):
            tester = "\n" * i + test_string
            with StringIO(tester) as fh:
                fh.lineno = 0
                fh.path = "foo"
                fh.name = "name"
                for card in input_syntax_reader.read_data(
                    fh, (6, 2, 0), recursion=True
                ):
                    pass
                assert montepy.input_parser.block_type.BlockType(i) == card.block_type

    def testCommentFormatInput(self):
        in_strs = ["c foo", "c bar"]
        card = montepy.input_parser.syntax_node.CommentNode(in_strs[0])
        output = card.format()
        answer = "c foo"
        str_answer = "COMMENT: c foo"
        assert repr(card) == str_answer
        assert "c foo" == str(card)
        assert len(answer) == len(output)
        for i, line in enumerate(output):
            assert answer[i] == line

    def testMessageFormatInput(self):
        answer = ["MESSAGE: foo", "bar", ""]
        card = montepy.input_parser.mcnp_input.Message(answer, ["foo", "bar"])
        str_answer = """MESSAGE:
foo
bar
"""
        assert str_answer == repr(card)
        assert "MESSAGE: 2 lines" == str(card)
        output = card.format_for_mcnp_input((6, 2, 0))
        assert len(answer) == len(output)
        for i, line in enumerate(output):
            assert answer[i] == line

    def testTitleFormatInput(self):
        card = montepy.input_parser.mcnp_input.Title(["foo"], "foo")
        answer = ["foo"]
        str_answer = "TITLE: foo"
        assert str(card) == str_answer
        output = card.format_for_mcnp_input((6, 2, 0))
        assert len(answer) == len(output)
        for i, line in enumerate(output):
            assert answer[i] == line

    def testReadInput(self):
        # TODO ensure comments are properly glued to right input
        generator = input_syntax_reader.read_input_syntax(
            MCNP_InputFile("tests/inputs/test.imcnp")
        )
        mcnp_in = montepy.input_parser.mcnp_input
        input_order = [mcnp_in.Message, mcnp_in.Title]
        input_order += [mcnp_in.Input] * 29
        for i, input in enumerate(generator):
            print(input.input_lines)
            print(input_order[i])
            assert isinstance(input, input_order[i])

    def testReadInputWithRead(self):
        generator = input_syntax_reader.read_input_syntax(
            MCNP_InputFile("tests/inputs/testRead.imcnp")
        )
        next(generator)  # skip title
        next(generator)  # skip read none
        next(generator)  # skip surfaces input
        next(generator)  # skip data mode input
        card = next(generator)
        answer = ["1 0 -1", "c"]
        assert answer == card.input_lines

    def testReadInputWithVertMode(self):
        generator = input_syntax_reader.read_input_syntax(
            MCNP_InputFile("tests/inputs/testVerticalMode.imcnp")
        )
        next(generator)
        next(generator)
        with pytest.raises(montepy.errors.UnsupportedFeature):
            next(generator)

    def testCardStringRepr(self):
        in_str = "1 0 -1"
        card = montepy.input_parser.mcnp_input.Input(
            [in_str], montepy.input_parser.block_type.BlockType.CELL
        )
        assert str(card) == "INPUT: BlockType.CELL"
        assert repr(card) == "INPUT: BlockType.CELL: ['1 0 -1']"

    def testDataInputNameParsing(self):
        tests = {
            "kcOde": {"prefix": "kcode", "number": None, "classifier": None},
            "M300": {"prefix": "m", "number": 300, "classifier": None},
            "IMP:N,P,E": {
                "prefix": "imp",
                "number": None,
                "classifier": [Particle.NEUTRON, Particle.PHOTON, Particle.ELECTRON],
            },
            "F1004:n,P": {
                "prefix": "f",
                "number": 1004,
                "classifier": [Particle.NEUTRON, Particle.PHOTON],
            },
        }
        for in_str, answer in tests.items():
            # Testing parsing the names
            print("in", in_str, "answer", answer)
            card = montepy.input_parser.mcnp_input.Input(
                [in_str], montepy.input_parser.block_type.BlockType.DATA
            )
            data_input = montepy.data_inputs.data_input.DataInput(card, fast_parse=True)
            assert data_input.prefix == answer["prefix"]
            if answer["number"]:
                assert data_input._input_number.value == answer["number"]
            if answer["classifier"]:
                assert sorted(data_input.particle_classifiers) == sorted(
                    answer["classifier"]
                )

    @pytest.mark.parametrize(
        "in_str, answer",
        [
            ("kcOde5", {"prefix": "kcode", "number": False, "classifier": 0}),
            ("M-300", {"prefix": "m", "number": True, "classifier": 0}),
            ("M", {"prefix": "m", "number": True, "classifier": 0}),
            ("f4m", {"prefix": "fm", "number": True, "classifier": 1}),
            ("IMP:N,P,E", {"prefix": "imp", "number": False, "classifier": 0}),
            ("IMP", {"prefix": "imp", "number": False, "classifier": 2}),
        ],
    )
    def test_data_name_enforce_bad(_, in_str, answer):
        with pytest.raises(montepy.errors.MalformedInputError):
            card = montepy.input_parser.mcnp_input.Input(
                [in_str], montepy.input_parser.block_type.BlockType.DATA
            )
            Fixture = DataInputTestFixture
            Fixture._class_prefix1 = answer["prefix"]
            Fixture._has_number1 = answer["number"]
            Fixture._has_classifier1 = answer["classifier"]
            card = Fixture(card)

    @pytest.mark.parametrize(
        "in_str, answer",
        [
            ("IMP:N,P,E", {"prefix": "imp", "number": False, "classifier": 2}),
            ("F1004:n,P", {"prefix": "f", "number": True, "classifier": 1}),
        ],
    )
    def test_dat_name_enforce_good(_, in_str, answer):
        card = montepy.input_parser.mcnp_input.Input(
            [in_str], montepy.input_parser.block_type.BlockType.DATA
        )
        Fixture = DataInputTestFixture
        Fixture._class_prefix1 = answer["prefix"]
        Fixture._has_number1 = answer["number"]
        Fixture._has_classifier1 = answer["classifier"]
        card = Fixture(card)

    @pytest.mark.parametrize(
        "version, line_number",
        [
            ((5, 1, 60), 80),
            ((6, 1, 0), 80),
            ((6, 2, 0), 128),
            ((6, 3, 0), 128),
            ((6, 3, 1), 128),
            ((6, 3, 3), 128),  # Test for newer not released versions
            ((7, 4, 0), 128),
        ],
    )
    def test_get_line_numbers(_, version, line_number):
        assert line_number == montepy.constants.get_max_line_length(version)
        with pytest.raises(montepy.errors.UnsupportedFeature):
            montepy.constants.get_max_line_length((5, 1, 38))

    def test_jump(self):
        jump = Jump()
        assert "J" == str(jump)
        jump2 = Jump()
        assert jump == jump2
        with pytest.raises(TypeError):
            bool(jump)

    def test_jump_and_a_hop(self):
        jump = Jump()
        # first you need to hop
        assert "j" == jump.lower()
        # then you need to skip
        assert "Jump" == jump.title()
        # before you can jump
        assert "J" == jump.upper()
        str(jump)
        repr(jump)


class TestClassifierNode:
    def test_classifier_init(self):
        classifier = syntax_node.ClassifierNode()
        assert None is (classifier.prefix)
        assert None is (classifier.number)
        assert None is (classifier.particles)
        assert None is (classifier.modifier)
        assert None is (classifier.padding)

    def test_classifier_setter(self):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("M", str)
        assert classifier.prefix.value == "M"
        classifier.number = syntax_node.ValueNode("124", int)
        assert classifier.number.value == 124
        classifier.modifier = syntax_node.ValueNode("*", str)
        assert classifier.modifier.value == "*"
        classifier.padding = syntax_node.PaddingNode(" ")
        assert len(classifier.padding.nodes) == 1

    def test_classifier_format(self):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("M", str)
        assert classifier.format() == "M"
        classifier.number = syntax_node.ValueNode("124", int)
        assert classifier.format() == "M124"
        classifier.modifier = syntax_node.ValueNode("*", str)
        assert classifier.format() == "*M124"
        classifier.padding = syntax_node.PaddingNode(" ")
        assert classifier.format() == "*M124 "
        str(classifier)
        repr(classifier)

    def test_classifier_comments(self):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("M", str)
        assert len(list(classifier.comments)) == 0
        classifier.padding = syntax_node.PaddingNode(" ")
        classifier.padding.append("$ hi", True)
        assert len(list(classifier.comments)) == 1


class TestParametersNode:
    @pytest.fixture
    def param(_):
        param = syntax_node.ParametersNode()
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = syntax_node.ValueNode("vol", str)
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode("1.0", float))
        param.append(
            syntax_node.SyntaxNode(
                "hi",
                {
                    "classifier": classifier,
                    "seperator": syntax_node.ValueNode("=", str),
                    "data": list_node,
                },
            )
        )
        return param

    def test_parameter_init(self):
        param = syntax_node.ParametersNode()
        assert isinstance(param.nodes, dict)

    def test_parameter_append(_, param):
        assert len(param.nodes) == 1
        with pytest.raises(ValueError):
            classifier = syntax_node.ClassifierNode()
            classifier.prefix = syntax_node.ValueNode("vol", str)
            param.append(syntax_node.SyntaxNode("foo", {"classifier": classifier}))

    def test_parameter_dict(_, param):
        assert param["vol"] == param.nodes["vol"]
        with pytest.raises(KeyError):
            param["hi"]
        assert "vol" in param

    def test_parameter_str(_, param):
        str(param)
        repr(param)
        param._pretty_str()

    def test_parameter_format(_, param):
        assert param.format() == "vol=1.0"

    def test_parameter_comments(_, param):
        assert len(list(param.comments)) == 0
        param["vol"]["data"][0].padding = syntax_node.PaddingNode(" ")
        param["vol"]["data"][0].padding.append("$ hi", True)
        assert len(list(param.comments)) == 1

    def test_parameter_trailing_comments(_, param):
        assert None is (param.get_trailing_comment())
        param._delete_trailing_comment()
        assert None is (param.get_trailing_comment())
        padding = syntax_node.PaddingNode("$ hi", True)
        param["vol"]["data"][0].padding = padding
        comment = param.get_trailing_comment()
        assert None is (comment)
        padding.append(syntax_node.CommentNode("c hi"), True)
        comment = param.get_trailing_comment()
        assert comment[0].contents == "hi"
        param._delete_trailing_comment()
        assert None is (param.get_trailing_comment())


class DataInputTestFixture(montepy.data_inputs.data_input.DataInputAbstract):
    _class_prefix1 = None
    _has_number1 = None
    _has_classifier1 = None

    def __init__(self, input_card=None):
        """
        :param input_card: the Card object representing this data input
        :type input_card: Input
        """
        super().__init__(input_card, fast_parse=True)

    def _class_prefix(self):
        return self._class_prefix1

    def _has_number(self):
        return self._has_number1

    def _has_classifier(self):
        return self._has_classifier1



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from hypothesis import given, note, strategies as st
from unittest import TestCase
import pytest

import montepy
from montepy.cell import Cell
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input


class TestCellClass(TestCase):
    def test_bad_init(self):
        with self.assertRaises(TypeError):
            Cell(5)

    # TODO test updating cell geometry once done
    def test_cell_validator(self):
        cell = Cell()
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            cell.format_for_mcnp_input((6, 2, 0))
        cell.mass_density = 5.0
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()
        del cell.mass_density

    # TODO test geometry stuff

    def test_number_setter(self):
        cell = Cell("1 0 2")
        cell.number = 5
        self.assertEqual(cell.number, 5)
        with self.assertRaises(TypeError):
            cell.number = "5"
        with self.assertRaises(ValueError):
            cell.number = -5

    def test_cell_density_setter(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        cell.mass_density = 1.5
        self.assertEqual(cell._density, 1.5)
        self.assertEqual(cell.mass_density, 1.5)
        self.assertFalse(cell.is_atom_dens)
        with self.assertRaises(AttributeError):
            _ = cell.atom_density
        cell.atom_density = 1.6
        self.assertEqual(cell._density, 1.6)
        self.assertEqual(cell.atom_density, 1.6)
        self.assertTrue(cell.is_atom_dens)
        with self.assertRaises(AttributeError):
            _ = cell.mass_density
        with self.assertRaises(TypeError):
            cell.atom_density = (5, True)
        with self.assertRaises(TypeError):
            cell.mass_density = "five"
        with self.assertRaises(ValueError):
            cell.atom_density = -1.5
        with self.assertRaises(ValueError):
            cell.mass_density = -5

    def test_cell_density_deleter(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        del cell.mass_density
        self.assertIsNone(cell.mass_density)
        cell.atom_density = 1.0
        del cell.atom_density
        self.assertIsNone(cell.atom_density)

    def test_cell_sorting(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell1 = Cell(card)
        in_str = "2 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell2 = Cell(card)
        test_sort = sorted([cell2, cell1])
        answer = [cell1, cell2]
        for i, cell in enumerate(test_sort):
            self.assertEqual(cell, answer[i])

    def test_cell_parameters_setting(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        params = {"FILL": "5"}
        cell.parameters = params
        self.assertEqual(params, cell.parameters)
        with self.assertRaises(TypeError):
            cell.parameters = []

    def test_cell_str(self):
        in_str = "1 1 0.5 2"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(str(cell), "CELL: 1, mat: 0, DENS: 0.5 atom/b-cm")
        self.assertEqual(
            repr(cell), "CELL: 1 \nVoid material \ndensity: 0.5 atom/b-cm\n"
        )

    def test_cell_paremeters_no_eq(self):
        in_str = f"1 0 -1 PWT 1.0"
        card = Input([in_str], BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(cell.parameters["PWT"]["data"][0].value, 1.0)


@pytest.mark.parametrize(
    "line, is_void, mat_number, density, atom_dens, parameters",
    [
        ("1 0 2", True, 0, None, None, None),
        ("1 1 -10 2", False, 1, 10.0, False, None),
        ("1 1 10 2", False, 1, 10.0, True, None),
        (
            "1 0 #2 u= 5 vol=20 trcl=5",
            True,
            0,
            None,
            None,
            {"TRCL": 5.0, "U": 5.0, "VOL": 20.0},
        ),
        (
            "1 1 -1.0 1000 (-3000:-4000:-5000)(-6000:-7000:8000) -2000 -9000",
            False,
            1,
            1,
            False,
            None,
        ),
    ],
)
def test_init(line, is_void, mat_number, density, atom_dens, parameters):
    # test like feature unsupported
    # tests void cell
    input = Input([line], BlockType.CELL)
    cell = Cell(input)
    assert cell.old_number == 1
    assert cell.number == 1
    if is_void:
        assert cell.material is None
    else:
        assert cell.old_mat_number != 0
        if atom_dens:
            assert cell.is_atom_dens
            assert cell.atom_density == pytest.approx(density)
        else:
            assert not cell.is_atom_dens
            assert cell.mass_density == pytest.approx(density)
    assert cell.old_mat_number == mat_number
    if not parameters:
        return
    for parameter, value in parameters.items():
        assert cell.parameters[parameter]["data"][0].value == pytest.approx(value)


def test_blank_num_init():
    cell = Cell(number=5)
    assert cell.number == 5
    with pytest.raises(TypeError):
        Cell(number="hi")
    with pytest.raises(ValueError):
        Cell(number=-1)


@pytest.mark.parametrize("line", ["foo", "foo bar", "1 foo", "1 1 foo"])
def test_malformed_init(line):
    with pytest.raises(montepy.errors.MalformedInputError):
        input = Input([line], BlockType.CELL)
        Cell(input)


@pytest.mark.parametrize("line", ["1 like 2"])
def test_malformed_init(line):
    with pytest.raises(montepy.errors.UnsupportedFeature):
        input = Input([line], BlockType.CELL)
        Cell(input)


@given(
    st.booleans(),
    st.booleans(),
    st.booleans(),
    st.integers(),
    st.integers(),
    st.integers(1),
    st.integers(1),
)
@pytest.mark.filterwarnings("ignore::montepy.errors.LineExpansionWarning")
def test_cell_clone(
    has_mat, clone_region, clone_material, start_num, step, other_start, other_step
):
    if has_mat:
        input = Input(["1 1 -0.5 2"], BlockType.CELL)
    else:
        input = Input(["1 0 2"], BlockType.CELL)
    surf = montepy.surfaces.surface.Surface()
    surf.number = 2
    mat = montepy.data_inputs.material.Material()
    mat.number = 1
    surfs = montepy.surface_collection.Surfaces([surf])
    mats = montepy.materials.Materials([mat])
    if other_start <= 0 or other_step <= 0:
        with pytest.raises(ValueError):
            mats.starting_number = other_start
            mats.step = other_step
    else:
        for collect in (surfs, mats):
            collect.starting_number = other_start
            collect.step = other_step
    cell = Cell(input)
    cell.update_pointers([], mats, surfs)
    problem = montepy.MCNP_Problem("foo")
    problem.surfaces = surfs
    problem.materials = mats
    for prob in {None, problem}:
        cell.link_to_problem(prob)
        if prob is not None:
            prob.cells.append(cell)
        if start_num <= 0 or step <= 0:
            with pytest.raises(ValueError):
                cell.clone(clone_material, clone_region, start_num, step)
            return
        new_cell = cell.clone(clone_material, clone_region, start_num, step)
        verify_internal_links(new_cell)
        verify_clone_format(new_cell)
        assert new_cell is not cell
        assert new_cell.number != 1
        if start_num != 1:
            assert new_cell.number == start_num
        else:
            assert new_cell.number == start_num + step
        # force it to use the step
        if prob is not None:
            new_cell2 = cell.clone(clone_material, clone_region, start_num, step)
            if start_num != 1:
                assert new_cell2.number == start_num + step
            else:
                assert new_cell2.number == start_num + step * 2
        for attr in {"_importance", "_volume", "_fill"}:
            assert getattr(cell, attr) is not getattr(new_cell, attr)
        for attr in {"mass_density"}:
            assert getattr(cell, attr) == getattr(new_cell, attr)
            if attr == "mass_density":
                attr = "density_node"
            assert getattr(cell, f"_{attr}") is not getattr(new_cell, f"_{attr}")
        assert cell.geometry is not new_cell.geometry
        if clone_region:
            assert list(cell.surfaces)[0] is not list(new_cell.surfaces)[0]
            assert list(new_cell.surfaces)[0].number != list(cell.surfaces)[0].number
        else:
            assert list(cell.surfaces)[0] is list(new_cell.surfaces)[0]
        if clone_material:
            if cell.material is None:
                assert new_cell.material is None
            else:
                assert cell.material is not new_cell.material
                assert new_cell.material.number != cell.material.number
        else:
            assert cell.material is new_cell.material


def verify_internal_links(cell):
    # verify _number is linked in tree
    assert cell._number is cell._tree["cell_num"]
    assert cell._old_mat_number is cell._tree["material"]["mat_number"]
    assert cell.geometry.node is cell._tree["geometry"].left
    assert cell.importance._tree is cell._tree["parameters"]["imp:n"]


def verify_clone_format(cell):
    surf = list(cell.surfaces)[0]
    old_num = surf.number
    num = 1000
    surf.number = num
    output = cell.format_for_mcnp_input((6, 3, 0))
    note(output)
    input = montepy.input_parser.mcnp_input.Input(
        output, montepy.input_parser.block_type.BlockType.CELL
    )
    new_cell = montepy.Cell(input)
    if cell.material is not None:
        mats = montepy.materials.Materials([cell.material])
    else:
        mats = []
    new_cell.update_pointers([], mats, montepy.surface_collection.Surfaces([surf]))
    for attr in {"number", "mass_density", "old_mat_number"}:
        assert getattr(cell, attr) == getattr(new_cell, attr)
    new_surf = list(new_cell.surfaces)[0]
    assert new_surf.number == num
    surf.number = old_num


def test_cell_clone_default():
    input = Input(["1 1 -0.5 2"], BlockType.CELL)
    cell = Cell(input)
    problem = montepy.MCNP_Problem("")
    problem.cells.append(cell)
    for prob in {problem, None}:
        cell.link_to_problem(prob)
        new_cell = cell.clone()
        assert new_cell.number != cell.number


@pytest.mark.parametrize(
    "args, error",
    [
        (("a", False, 1, 1), TypeError),
        ((False, "b", 1, 1), TypeError),
        ((False, False, "c", 1), TypeError),
        ((False, False, 1, "d"), TypeError),
        ((False, False, -1, 1), ValueError),
        ((False, False, 0, 1), ValueError),
        ((False, False, 1, 0), ValueError),
        ((False, False, 1, -1), ValueError),
    ],
)
def test_cell_clone_bad(args, error):
    input = Input(["1 0 2"], BlockType.CELL)
    cell = Cell(input)
    surf = montepy.surfaces.surface.Surface()
    surf.number = 2
    surfs = montepy.surface_collection.Surfaces([surf])
    cell.update_pointers([], [], surfs)
    with pytest.raises(error):
        cell.clone(*args)


def test_bad_setattr():
    cell = montepy.Cell()
    with pytest.raises(AttributeError):
        cell.nuber = 5
    cell._nuber = 5
    assert cell._nuber == 5



import gc

import time
import tracemalloc

tracemalloc.start()

import montepy

FAIL_THRESHOLD = 40
MEMORY_FRACTION = 0.50

starting_mem = tracemalloc.get_traced_memory()[0]
print(f"starting memory with montepy. {starting_mem/1024/1024} MB")
start = time.time()

problem = montepy.read_input("benchmark/big_model.imcnp")

stop = time.time()

problem_mem = tracemalloc.get_traced_memory()[0]
print(f"Took {stop - start} seconds")
print(f"Memory usage report: {problem_mem/1024/1024} MB")
del problem
gc.collect()
ending_mem = tracemalloc.get_traced_memory()[0]
print(f"Memory usage report after GC: {ending_mem/1024/1024} MB")

if (stop - start) > FAIL_THRESHOLD:
    raise RuntimeError(
        f"Benchmark took too long to complete. It must be faster than: {FAIL_THRESHOLD} s."
    )

prob_gc_mem = problem_mem - ending_mem
prob_actual_mem = problem_mem - starting_mem
gc_ratio = prob_gc_mem / prob_actual_mem
print(f"{gc_ratio:.2%} of the problem's memory was garbage collected.")
if (prob_gc_mem / prob_actual_mem) < MEMORY_FRACTION:
    raise RuntimeError(
        f"Benchmark had too many memory leaks. Only {gc_ratio:.2%} of the memory was collected."
    )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import collections
import copy
import math
import warnings
from montepy.data_inputs.cell_modifier import CellModifierInput, InitInput
from montepy.errors import *
from montepy.constants import DEFAULT_VERSION, rel_tol, abs_tol
from montepy.input_parser import syntax_node
from montepy.mcnp_object import MCNP_Object
from montepy.particle import Particle
from montepy.utilities import *
import numbers

#
# ********************* Developer Notes ********************************************
#
# How Importance handles syntax trees is complicated.
# One object holds information for N particle types.
# This can be associated with between 1 and N syntax trees (imp:n,p v. imp:n imp:p)
#
# Variables
#
# * _tree      : the syntax tree from parsing. Only used on initial parsing
# * _real_tree : holds unique trees for every particle type. This is used in data block formatting.
# * _particle_importances : a dictionary of ParameterNodes that maps a particle to it's ParameterNode
# * _part_combos : a list of ParticleNode that show which particles were combined on the original input


class Importance(CellModifierInput):
    """A data input that sets the importance for a cell(s).

    Parameters
    ----------
    input : Union[Input, str]
        the Input object representing this data input
    in_cell_block : bool
        if this card came from the cell block of an input file.
    key : str
        the key from the key-value pair in a cell
    value : SyntaxNode
        the value syntax tree from the key-value pair in a cell
    """

    def __init__(
        self,
        input: InitInput = None,
        in_cell_block: bool = False,
        key: str = None,
        value: syntax_node.SyntaxNode = None,
    ):
        self._particle_importances = {}
        self._real_tree = {}
        self._part_combos = []
        super().__init__(input, in_cell_block, key, value)
        if self.in_cell_block:
            if key:
                val = value["data"]
                if isinstance(val, syntax_node.ListNode):
                    val = value["data"][0]
                if val.type != float or val.value < 0:
                    raise ValueError(
                        f"Cell importance must be a number ≥ 0. {val.value} was given"
                    )
                self._part_combos.append(self.particle_classifiers)
                for particle in self.particle_classifiers:
                    self._particle_importances[particle] = value
        elif input:
            values = []
            for node in self._tree["data"]:
                try:
                    value = node.value
                    assert value >= 0
                    values.append(node)
                except (AttributeError, AssertionError) as e:
                    raise MalformedInputError(
                        input, f"Importances must be ≥ 0 value: {node} given"
                    )
            self._part_combos.append(self.particle_classifiers)
            for particle in self.particle_classifiers:
                self._particle_importances[particle] = copy.deepcopy(self._tree)
                self._real_tree[particle] = copy.deepcopy(self._tree)

    def _generate_default_cell_tree(self, particle=None):
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = self._generate_default_node(
            str, self._class_prefix().upper(), None
        )
        if particle is None:
            particles = syntax_node.ParticleNode("imp particle", "n")
            particle = Particle.NEUTRON
        else:
            particles = syntax_node.ParticleNode("imp particle", particle.value.lower())
        if self._problem:
            particles.particles = self._problem.mode.particles
        classifier.particles = particles
        list_node = syntax_node.ListNode("imp data")
        list_node.append(self._generate_default_node(float, 0.0))
        tree = syntax_node.SyntaxNode(
            "Importance",
            {
                "classifier": classifier,
                "seperator": self._generate_default_node(str, "=", padding=None),
                "data": list_node,
            },
        )
        self._tree = tree
        self._particle_importances[particle] = tree

    @property
    def _tree_value(self, particle):
        pass

    @staticmethod
    def _class_prefix():
        return "imp"

    @staticmethod
    def _has_number():
        return False

    @staticmethod
    def _has_classifier():
        return 2

    @property
    def has_information(self):
        if self.in_cell_block:
            return True

    def merge(self, other):
        if not isinstance(other, type(self)):
            raise TypeError("Can only be merged with other Importance object")
        if self.in_cell_block != other.in_cell_block:
            raise ValueError("Can not mix cell-level and data-level Importance objects")
        if other.set_in_cell_block:
            self._set_in_cell_block = True
        for particle in other:
            if particle not in self:
                self._particle_importances[particle] = other._particle_importances[
                    particle
                ]
                # keep original formatting external to data cleared out
                if not self.in_cell_block:
                    self._real_tree[particle] = self._particle_importances[particle]
            else:
                raise MalformedInputError(
                    other._input,
                    "Cannot have two importance inputs for the same particle type",
                )

    def __iter__(self):
        return iter(self._particle_importances.keys())

    def __contains__(self, value):
        return value in self._particle_importances

    def __getitem__(self, particle):
        if not isinstance(particle, Particle):
            raise TypeError("Key must be a particle")
        self._check_particle_in_problem(particle)
        try:
            val = self._particle_importances[particle]["data"][0]
            return val.value
        except KeyError:
            return 0.0

    def __setitem__(self, particle, value):
        if not isinstance(particle, Particle):
            raise TypeError("Key must be a particle")
        self._check_particle_in_problem(particle)
        if not isinstance(value, numbers.Number):
            raise TypeError("importance must be a number")
        if value < 0:
            raise ValueError("importance must be ≥ 0")
        if particle not in self._particle_importances:
            self._generate_default_cell_tree(particle)
        self._particle_importances[particle]["data"][0].value = value

    def __delitem__(self, particle):
        if not isinstance(particle, Particle):
            raise TypeError("Key must be a particle")
        del self._particle_importances[particle]

    def __str__(self):
        if not self.in_cell_block and self._problem is None:
            return " ".join(self.input_lines)
        return "".join(self.format_for_mcnp_input(DEFAULT_VERSION))

    def __repr__(self):
        return (
            f"Importance: in_cell_block: {self.in_cell_block},"
            f" set_in_cell_block {self.set_in_cell_block},"
            f"\n{self._particle_importances}"
        )

    def push_to_cells(self):
        if self._problem and not self.in_cell_block:
            self._check_redundant_definitions()
            for particle in self._particle_importances:
                if not self._particle_importances[particle]:
                    continue
                for i, cell in enumerate(self._problem.cells):
                    value = self._particle_importances[particle]["data"][i]
                    # force generating the default tree
                    cell.importance[particle] = value.value
                    # replace default ValueNode with actual valueNode
                    tree = cell.importance._particle_importances[particle]
                    tree.nodes["classifier"] = copy.deepcopy(
                        self._particle_importances[particle]["classifier"]
                    )
                    tree["classifier"].padding = None
                    data = tree["data"]
                    data.nodes.pop()
                    data.nodes.append(value)

    def _format_tree(self):
        if self.in_cell_block:
            particles_printed = set()
            ret = ""
            for particle in self:
                if particle in particles_printed:
                    continue
                other_particles = self._particle_importances[particle][
                    "classifier"
                ].particles
                to_remove = set()
                for other_part in other_particles:
                    if other_part != particle:
                        if math.isclose(
                            self[particle],
                            self[other_part],
                            rel_tol=rel_tol,
                            abs_tol=abs_tol,
                        ):
                            particles_printed.add(other_part)
                        else:
                            to_remove.add(other_part)
                for removee in to_remove:
                    other_particles.remove(removee)
                ret += self._particle_importances[particle].format()
                particles_printed.add(particle)
            return ret
        else:
            printed_parts = set()
            ret = []
            for particle, tree in self._real_tree.items():
                if particle in printed_parts:
                    continue
                printed_parts |= tree["classifier"].particles.particles
                ret.append(tree.format())
            return "\n".join(ret)

    @property
    def all(self):
        """Setter for setting importance for all particle types in the problem at once.

        Parameters
        ----------
        importance : float
            the importance to set all particles to.

        Returns
        -------
        None
            None
        """
        return None

    @all.setter
    def all(self, value):
        if not isinstance(value, numbers.Number):
            raise TypeError("All importance must be a float")
        value = float(value)
        if value < 0.0:
            raise ValueError("Importance must be ≥ 0.0")
        if self._problem:
            for particle in self._problem.mode:
                self._particle_importances[particle]["data"][0].value = value

    def _clear_data(self):
        if not self.in_cell_block:
            self._particle_importances = {}

    def _check_particle_in_problem(self, particle_type):
        if self._problem:
            if particle_type not in self._problem.mode:
                warnings.warn(
                    f"Particle type: {particle_type} not included in problem mode.",
                    ParticleTypeNotInProblem,
                )

    def _collect_new_values(self):
        new_vals = collections.defaultdict(list)
        particle_pairings = collections.defaultdict(set)
        for particle in self._problem.mode.particles:
            for cell in self._problem.cells:
                try:
                    tree = cell.importance._particle_importances[particle]
                except KeyError:
                    raise NotImplementedError(
                        f"Importance data not available for cell {cell.number} for particle: "
                        f"{particle}, though it is in the problem, and default importance logic "
                        "is not yet implemented in MontePy."
                    )
                new_vals[particle].append(tree["data"][0])
                if len(particle_pairings[particle]) == 0:
                    particle_pairings[particle] = tree["classifier"].particles.particles
                else:
                    particle_pairings[particle] &= tree[
                        "classifier"
                    ].particles.particles
        return self._try_combine_values(new_vals, particle_pairings)

    def _update_values(self, in_middle=False):
        if self.in_cell_block:
            self._update_cell_values()
            length = len(self._particle_importances)
            for i, tree in enumerate(self._particle_importances.values()):
                if i < length - 1:
                    edge = tree["data"][-1]
                    if isinstance(edge, syntax_node.ValueNode) and edge.padding is None:
                        edge.padding = syntax_node.PaddingNode(" ")
        else:
            new_vals = self._collect_new_values()
            for part_set, data in new_vals.items():
                for particle in part_set:
                    if particle not in self._real_tree:
                        tree = _generate_default_data_tree(particle)
                        self._real_tree[particle] = tree
                        padding = tree["classifier"].padding
                        tree.nodes["classifier"] = copy.deepcopy(
                            next(
                                iter(self._problem.cells)
                            ).importance._particle_importances[particle]["classifier"]
                        )
                        tree["classifier"].padding = padding
                    tree = self._real_tree[particle]
                    tree["classifier"].particles.particles = set(part_set)
                    tree["data"].update_with_new_values(data)

    def _try_combine_values(self, new_vals, particle_pairings):
        covered_parts = set()
        ret = {}
        for particle, pairings in particle_pairings.items():
            if particle in covered_parts:
                continue
            gold = new_vals[particle]
            matching_parts = {particle}
            covered_parts.add(particle)
            for test_part in pairings:
                if test_part == particle or test_part in covered_parts:
                    continue
                test_vals = new_vals[test_part]
                matches = True
                for gold_val, test_val in zip(gold, test_vals):
                    if not math.isclose(
                        gold_val.value,
                        test_val.value,
                        rel_tol=rel_tol,
                        abs_tol=abs_tol,
                    ):
                        matches = False
                        break
                if matches:
                    matching_parts.add(test_part)
                    covered_parts.add(test_part)
            ret[frozenset(matching_parts)] = gold
        return ret

    def _update_cell_values(self):
        pass

    @property
    def trailing_comment(self):
        """The trailing comments and padding of an input.

        Generally this will be blank as these will be moved to be a leading comment for the next input.

        Returns
        -------
        list
            the trailing ``c`` style comments and intermixed padding
            (e.g., new lines)
        """
        last_tree = list(self._real_tree.values())[-1]
        if last_tree:
            return last_tree.get_trailing_comment()

    def _delete_trailing_comment(self):
        for part, tree in reversed(self._real_tree.items()):
            tree._delete_trailing_comment()
            self.__delete_common_trailing(part)
            break

    def __delete_common_trailing(self, part):
        to_delete = {part}
        for combo_set in self._part_combos:
            if part in combo_set:
                to_delete |= combo_set
        if self._in_cell_block:
            for part in to_delete:
                self._particle_importances[part]["data"]._delete_trailing_comment()
        else:
            for part in to_delete:
                self._real_tree[part]._delete_trailing_comment()

    def _grab_beginning_comment(self, new_padding, last_obj=None):
        last_tree = None
        last_padding = None
        if self._in_cell_block:
            if not isinstance(last_obj, Importance):
                for part, tree in self._particle_importances.items():
                    if last_padding is not None and last_tree is not None:
                        last_tree._grab_beginning_comment(last_padding)
                        self.__delete_common_trailing(part)
                    last_padding = tree.get_trailing_comment()
                    last_tree = tree
                if new_padding:
                    next(iter(self._particle_importances.values()))[
                        "start_pad"
                    ]._grab_beginning_comment(new_padding)
        else:
            # if not inside a block of importances
            if not isinstance(last_obj, Importance):
                for part, tree in self._real_tree.items():
                    if tree.get_trailing_comment() == last_padding:
                        continue
                    if last_padding is not None and last_tree is not None:
                        last_tree._grab_beginning_comment(last_padding)
                        self.__delete_common_trailing(part)
                    last_padding = tree.get_trailing_comment()
                    last_tree = tree
                if new_padding:
                    next(iter(self._real_tree.values()))[
                        "start_pad"
                    ]._grab_beginning_comment(new_padding)
            # otherwise keep it as is inside the block
            else:
                list(self._real_tree.values())[-1]["start_pad"]._grab_beginning_comment(
                    new_padding
                )


def _generate_default_data_tree(particle):
    list_node = syntax_node.ListNode("number sequence")
    list_node.append(syntax_node.ValueNode(None, float))
    classifier = syntax_node.ClassifierNode()
    classifier.prefix = syntax_node.ValueNode("IMP", str)
    classifier.padding = syntax_node.PaddingNode(" ")
    classifier.particles = syntax_node.ParticleNode(
        "IMP_particles", f":{particle.value}"
    )
    classifier.particles.particles = {particle}
    return syntax_node.SyntaxNode(
        "IMP",
        {
            "start_pad": syntax_node.PaddingNode(),
            "classifier": classifier,
            "keyword": syntax_node.ValueNode(None, str, None),
            "data": list_node,
        },
    )


def __create_importance_getter(particle_type):
    def closure(obj):
        return obj[particle_type]

    return closure


def __create_importance_setter(particle_type):
    def closure(obj, value):
        obj[particle_type] = value

    return closure


def __create_importance_deleter(particle_type):
    def closure(obj):
        del obj[particle_type]

    return closure


def __create_particle_imp_doc(particle_type):
    return f"""Importance for particles of type *{particle_type.name.lower()}*

Can only be set if this particle is used in the problem mode.

Parameters
----------
importance: float 
    The importance to set this to.

Returns
-------
float
    the importance for the particle type. If not set, defaults to 0.
"""


def __setup_importances():
    for particle in Particle:
        getter = __create_importance_getter(particle)
        setter = __create_importance_setter(particle)
        deleter = __create_importance_deleter(particle)
        doc = __create_particle_imp_doc(particle)
        prop = property(getter, setter, deleter, doc=doc)
        setattr(Importance, particle.name.lower(), prop)


__setup_importances()



import pstats
from pstats import SortKey

stats = pstats.Stats("prof/combined.prof")
stats.sort_stats(SortKey.CUMULATIVE, SortKey.TIME).print_stats("montepy", 50)
stats.sort_stats(SortKey.CUMULATIVE, SortKey.TIME).print_stats("sly", 20)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum


class Operator(Enum):
    """Enumeration of the allowed geometry set logic."""

    INTERSECTION = "*"
    """Represents the intersection of sets."""
    UNION = ":"
    """Represents the union of sets."""
    COMPLEMENT = "#"
    """Represents the complement of a set."""
    _SHIFT = ">"
    """Internal operator essentially equivalent to No-op.

    This is used to properly handle some leaf nodes.
    """
    GROUP = "()"
    """Grouping operator that represents parentheses."""
